# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any, Callable, Literal, Optional, Tuple, Union

import jax
import jax.numpy as jnp

from ott import utils
from ott.geometry import costs, geometry
from ott.math import utils as mu

__all__ = ["LRCGeometry", "LRKGeometry"]


@jax.tree_util.register_pytree_node_class
class LRCGeometry(geometry.Geometry):
  """Geometry whose cost is defined by product of two low-rank matrices.

  Implements geometries that are defined as low rank products, i.e. for which
  there exists two matrices :math:`A` and :math:`B` of :math:`r` columns such
  that the cost of the geometry equals :math:`AB^T`. Apart from being faster to
  apply to a vector, these geometries are characterized by the fact that adding
  two such geometries should be carried out by concatenating factors, i.e.
  if :math:`C = AB^T` and :math:`D = EF^T` then :math:`C + D = [A,E][B,F]^T`

  Args:
    cost_1: Array of shape ``[num_a, r]``.
    cost_2: Array of shape ``[num_b, r]``.
    bias: constant added to entire cost matrix.
    scale: Value used to rescale the factors of the low-rank geometry.
    scale_cost: option to rescale the cost matrix. Implemented scalings are
      'max_bound', 'mean' and 'max_cost'. Alternatively, a float
      factor can be given to rescale the cost such that
      ``cost_matrix /= scale_cost``.
    kwargs: keyword arguments for :class:`~ott.geometry.geometry.Geometry`.
  """

  def __init__(
      self,
      cost_1: jnp.ndarray,
      cost_2: jnp.ndarray,
      bias: float = 0.0,
      scale_factor: float = 1.0,
      scale_cost: Union[float, Literal["mean", "max_bound", "max_cost"]] = 1.0,
      **kwargs: Any,
  ):
    super().__init__(**kwargs)
    self._cost_1 = cost_1
    self._cost_2 = cost_2
    self._bias = bias
    self._scale_factor = scale_factor
    self._scale_cost = scale_cost

  @property
  def cost_1(self) -> jnp.ndarray:
    """First factor of the :attr:`cost_matrix`."""
    scale_factor = jnp.sqrt(self._scale_factor * self.inv_scale_cost)
    return scale_factor * self._cost_1

  @property
  def cost_2(self) -> jnp.ndarray:
    """Second factor of the :attr:`cost_matrix`."""
    scale_factor = jnp.sqrt(self._scale_factor * self.inv_scale_cost)
    return scale_factor * self._cost_2

  @property
  def bias(self) -> float:
    """Constant offset added to the entire :attr:`cost_matrix`."""
    return self._bias * self.inv_scale_cost

  @property
  def cost_rank(self) -> int:  # noqa: D102
    return self._cost_1.shape[1]

  @property
  def cost_matrix(self) -> jnp.ndarray:
    """Materialize the cost matrix."""
    return jnp.matmul(self.cost_1, self.cost_2.T) + self.bias

  @property
  def shape(self) -> Tuple[int, int]:  # noqa: D102
    return self._cost_1.shape[0], self._cost_2.shape[0]

  @property
  def is_symmetric(self) -> bool:  # noqa: D102
    return (
        self._cost_1.shape[0] == self._cost_2.shape[0] and
        jnp.all(self._cost_1 == self._cost_2)
    )

  @property
  def inv_scale_cost(self) -> float:  # noqa: D102
    if isinstance(self._scale_cost, (int, float, jax.Array)):
      return 1.0 / self._scale_cost
    self = self._masked_geom()
    if self._scale_cost == "max_bound":
      x_norm = self._cost_1[:, 0].max()
      y_norm = self._cost_2[:, 1].max()
      max_bound = x_norm + y_norm + 2.0 * jnp.sqrt(x_norm * y_norm)
      return 1.0 / (max_bound + self._bias)
    if self._scale_cost == "mean":
      a, b = self._n_normed_ones, self._m_normed_ones
      mean = jnp.linalg.multi_dot([a, self._cost_1, self._cost_2.T, b])
      return 1.0 / (mean + self._bias)
    if self._scale_cost == "max_cost":
      return 1.0 / self._max_cost_matrix
    raise ValueError(f"Scaling {self._scale_cost} not implemented.")

  def apply_square_cost(self, arr: jnp.ndarray, axis: int = 0) -> jnp.ndarray:
    """Apply elementwise-square of cost matrix to array (vector or matrix)."""
    (n, m), r = self.shape, self.cost_rank
    # When applying square of a LRCGeometry, one can either elementwise square
    # the cost matrix, or instantiate an augmented (rank^2) LRCGeometry
    # and apply it. First is O(nm), the other is O((n+m)r^2).
    if n * m < (n + m) * r ** 2:  # better use regular apply
      return super().apply_square_cost(arr, axis)

    new_cost_1 = self.cost_1[:, :, None] * self.cost_1[:, None, :]
    new_cost_2 = self.cost_2[:, :, None] * self.cost_2[:, None, :]
    return LRCGeometry(
        cost_1=new_cost_1.reshape((n, r ** 2)),
        cost_2=new_cost_2.reshape((m, r ** 2))
    ).apply_cost(arr, axis)

  def _apply_cost_to_vec(
      self,
      vec: jnp.ndarray,
      axis: int = 0,
      fn: Optional[Callable[[jnp.ndarray], jnp.ndarray]] = None,
      is_linear: bool = False,
  ) -> jnp.ndarray:
    """Apply [num_a, num_b] fn(cost) (or transpose) to vector.

    Args:
      vec: jnp.ndarray [num_a,] ([num_b,] if axis=1) vector
      axis: axis on which the reduction is done.
      fn: function optionally applied to cost matrix element-wise, before the
        doc product
      is_linear: Whether ``fn`` is a linear function to enable efficient
        implementation.

    Returns:
      A jnp.ndarray corresponding to cost x vector
    """
    if fn is None or is_linear:
      return self._apply_cost_to_vec_fast(vec, axis, fn=fn)
    return super()._apply_cost_to_vec(vec, axis, fn=fn, is_linear=is_linear)

  def _apply_cost_to_vec_fast(
      self,
      vec: jnp.ndarray,
      axis: int = 0,
      fn: Optional[Callable[[jnp.ndarray], jnp.ndarray]] = None,
  ) -> jnp.ndarray:
    c1, c2 = (self.cost_1,
              self.cost_2) if axis == 1 else (self.cost_2, self.cost_1)
    bias = self.bias
    if fn is not None:
      c2, bias = fn(c2), fn(bias)
    out = jnp.linalg.multi_dot([c1, c2.T, vec])
    return out + bias * jnp.sum(vec) * jnp.ones_like(out)

  @property
  def _max_cost_matrix(self) -> jnp.ndarray:
    fn = utils.batched_vmap(
        lambda c1, c2: jnp.max(c1 @ c2.T), batch_size=1024, in_axes=(0, None)
    )
    return jnp.max(fn(self._cost_1, self._cost_2)) + self._bias

  def to_LRCGeometry(
      self,
      rank: int = 0,
      tol: float = 1e-2,
      rng: Optional[jax.Array] = None,
      scale: float = 1.0,
  ) -> "LRCGeometry":
    """Return self."""
    del rank, tol, rng, scale
    return self

  @property
  def can_LRC(self):  # noqa: D102
    return True

  def subset(  # noqa: D102
      self, src_ixs: Optional[jnp.ndarray], tgt_ixs: Optional[jnp.ndarray],
      **kwargs: Any
  ) -> "LRCGeometry":

    def subset_fn(
        arr: Optional[jnp.ndarray],
        ixs: Optional[jnp.ndarray],
    ) -> jnp.ndarray:
      return arr if arr is None or ixs is None else arr[ixs, ...]

    return self._mask_subset_helper(
        src_ixs, tgt_ixs, fn=subset_fn, propagate_mask=True, **kwargs
    )

  def mask(  # noqa: D102
      self,
      src_mask: Optional[jnp.ndarray],
      tgt_mask: Optional[jnp.ndarray],
      mask_value: float = 0.0,
  ) -> "LRCGeometry":

    def mask_fn(
        arr: Optional[jnp.ndarray],
        mask: Optional[jnp.ndarray],
    ) -> Optional[jnp.ndarray]:
      if arr is None or mask is None:
        return arr
      return jnp.where(mask[:, None], arr, mask_value)

    src_mask = self._normalize_mask(src_mask, self.shape[0])
    tgt_mask = self._normalize_mask(tgt_mask, self.shape[1])
    return self._mask_subset_helper(
        src_mask, tgt_mask, fn=mask_fn, propagate_mask=False
    )

  def _mask_subset_helper(
      self,
      src_ixs: Optional[jnp.ndarray],
      tgt_ixs: Optional[jnp.ndarray],
      *,
      fn: Callable[[Optional[jnp.ndarray], Optional[jnp.ndarray]],
                   Optional[jnp.ndarray]],
      propagate_mask: bool,
      **kwargs: Any,
  ) -> "LRCGeometry":
    (c1, c2, src_mask, tgt_mask, *children), aux_data = self.tree_flatten()
    c1 = fn(c1, src_ixs)
    c2 = fn(c2, tgt_ixs)
    if propagate_mask:
      src_mask = self._normalize_mask(src_mask, self.shape[0])
      tgt_mask = self._normalize_mask(tgt_mask, self.shape[1])
      src_mask = fn(src_mask, src_ixs)
      tgt_mask = fn(tgt_mask, tgt_ixs)

    aux_data = {**aux_data, **kwargs}
    return type(self).tree_unflatten(
        aux_data, [c1, c2, src_mask, tgt_mask] + children
    )

  def __add__(self, other: "LRCGeometry") -> "LRCGeometry":
    if not isinstance(other, LRCGeometry):
      return NotImplemented
    return LRCGeometry(
        cost_1=jnp.concatenate((self.cost_1, other.cost_1), axis=1),
        cost_2=jnp.concatenate((self.cost_2, other.cost_2), axis=1),
        bias=self._bias + other._bias,
        # already included in `cost_{1,2}`
        scale_factor=1.0,
        scale_cost=1.0,
    )

  @property
  def dtype(self) -> jnp.dtype:  # noqa: D102
    return self._cost_1.dtype

  def tree_flatten(self):  # noqa: D102
    return (
        self._cost_1,
        self._cost_2,
        self._src_mask,
        self._tgt_mask,
        self._epsilon_init,
        self._bias,
        self._scale_factor,
    ), {
        "scale_cost": self._scale_cost,
    }

  @classmethod
  def tree_unflatten(cls, aux_data, children):  # noqa: D102
    c1, c2, src_mask, tgt_mask, epsilon, bias, scale_factor = children
    return cls(
        c1,
        c2,
        bias=bias,
        scale_factor=scale_factor,
        epsilon=epsilon,
        src_mask=src_mask,
        tgt_mask=tgt_mask,
        **aux_data
    )


@jax.tree_util.register_pytree_node_class
class LRKGeometry(geometry.Geometry):
  """Low-rank kernel geometry.

  .. note::
    This constructor is not meant to be called by the user,
    please use the :meth:`from_pointcloud` method instead.

  Args:
    k1: Array of shape ``[num_a, r]`` with positive features.
    k2: Array of shape ``[num_b, r]`` with positive features.
    epsilon: Epsilon regularization.
    kwargs: Keyword arguments for :class:`~ott.geometry.geometry.Geometry`.
  """

  def __init__(
      self,
      k1: jnp.ndarray,
      k2: jnp.ndarray,
      epsilon: Optional[float] = None,
      **kwargs: Any
  ):
    super().__init__(epsilon=epsilon, relative_epsilon=False, **kwargs)
    self.k1 = k1
    self.k2 = k2

  @classmethod
  def from_pointcloud(
      cls,
      x: jnp.ndarray,
      y: jnp.ndarray,
      *,
      kernel: Literal["gaussian", "arccos"],
      rank: int = 100,
      std: float = 1.0,
      n: int = 1,
      rng: Optional[jax.Array] = None
  ) -> "LRKGeometry":
    r"""Low-rank kernel approximation :cite:`scetbon:20`.

    Args:
      x: Array of shape ``[n, d]``.
      y: Array of shape ``[m, d]``.
      kernel: Type of the kernel to approximate.
      rank: Rank of the approximation.
      std: Depending on the ``kernel`` approximation:

        - ``'gaussian'`` - scale of the Gibbs kernel.
        - ``'arccos'`` - standard deviation of the random projections.
      n: Order of the arc-cosine kernel, see :cite:`cho:09` for reference.
      rng: Random key used for seeding.

    Returns:
      Low-rank kernel geometry.
    """
    rng = utils.default_prng_key(rng)
    if kernel == "gaussian":
      r = jnp.maximum(
          jnp.linalg.norm(x, axis=-1).max(),
          jnp.linalg.norm(y, axis=-1).max()
      )
      k1 = _gaussian_kernel(rng, x, rank, eps=std, R=r)
      k2 = _gaussian_kernel(rng, y, rank, eps=std, R=r)
      eps = std
    elif kernel == "arccos":
      k1 = _arccos_kernel(rng, x, rank, n=n, std=std)
      k2 = _arccos_kernel(rng, y, rank, n=n, std=std)
      eps = 1.0
    else:
      raise NotImplementedError(kernel)

    return cls(k1, k2, epsilon=eps)

  def apply_kernel(  # noqa: D102
      self,
      vec: jnp.ndarray,
      eps: Optional[float] = None,
      axis: int = 0,
  ) -> jnp.ndarray:
    if axis == 0:
      return self.k2 @ (self.k1.T @ vec)
    return self.k1 @ (self.k2.T @ vec)

  @property
  def kernel_matrix(self) -> jnp.ndarray:  # noqa: D102
    return self.k1 @ self.k2.T

  @property
  def cost_matrix(self) -> jnp.ndarray:  # noqa: D102
    eps = jnp.finfo(self.dtype).tiny
    return -self.epsilon * jnp.log(self.kernel_matrix + eps)

  @property
  def rank(self) -> int:  # noqa: D102
    return self.k1.shape[1]

  @property
  def shape(self) -> Tuple[int, int]:  # noqa: D102
    return self.k1.shape[0], self.k2.shape[0]

  @property
  def dtype(self) -> jnp.dtype:  # noqa: D102
    return self.k1.dtype

  def transport_from_potentials(
      self, f: jnp.ndarray, g: jnp.ndarray
  ) -> jnp.ndarray:
    """Not implemented."""
    raise ValueError("Not implemented.")

  def tree_flatten(self):  # noqa: D102
    return [self.k1, self.k2, self._epsilon_init], {}

  @classmethod
  def tree_unflatten(cls, aux_data, children):  # noqa: D102
    return cls(*children, **aux_data)


def _gaussian_kernel(
    rng: jax.Array,
    x: jnp.ndarray,
    n_features: int,
    eps: float,
    R: jnp.ndarray,
) -> jnp.ndarray:
  _, d = x.shape
  cost_fn = costs.SqEuclidean()

  y = (R ** 2) / (eps * d)
  q = y / (2.0 * mu.lambertw(y))
  sigma = jnp.sqrt(q * eps * 0.25)

  u = jax.random.normal(rng, shape=(n_features, d)) * sigma
  cost = cost_fn.all_pairs(x, u)
  norm_u = cost_fn.norm(u)

  tmp = -2.0 * (cost / eps) + (norm_u / (eps * q))
  phi = (2 * q) ** (d / 4) * jnp.exp(tmp)

  return (1.0 / jnp.sqrt(n_features)) * phi


def _arccos_kernel(
    rng: jax.Array,
    x: jnp.ndarray,
    n_features: int,
    n: int,
    std: float = 1.0,
    kappa: float = 1e-6,
) -> jnp.ndarray:
  n_points, d = x.shape
  c = jnp.sqrt(2) * (std ** (d / 2))

  u = jax.random.normal(rng, shape=(n_features, d)) * std
  tmp = -(1 / 4) * jnp.sum(u ** 2, axis=-1) * (1.0 - (1.0 / (std ** 2)))
  phi = c * (jnp.maximum(0.0, (x @ u.T)) ** n) * jnp.exp(tmp)

  return jnp.c_[(1.0 / jnp.sqrt(n_features)) * phi,
                jnp.full((n_points,), fill_value=kappa)]



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any, Optional, Union

import jax.numpy as jnp

from ott.geometry import geometry, pointcloud
from ott.problems.linear import linear_problem
from ott.solvers.linear import sinkhorn, sinkhorn_lr, univariate

__all__ = ["solve", "solve_univariate"]


def solve(
    geom: geometry.Geometry,
    a: Optional[jnp.ndarray] = None,
    b: Optional[jnp.ndarray] = None,
    tau_a: float = 1.0,
    tau_b: float = 1.0,
    rank: int = -1,
    **kwargs: Any
) -> Union[sinkhorn.SinkhornOutput, sinkhorn_lr.LRSinkhornOutput]:
  """Solve linear regularized OT problem using Sinkhorn iterations.

  Args:
    geom: The ground geometry of the linear problem.
    a: The first marginal. If :obj:`None`, it will be uniform.
    b: The second marginal. If :obj:`None`, it will be uniform.
    tau_a: If :math:`< 1`, defines how much unbalanced the problem is
      on the first marginal.
    tau_b: If :math:`< 1`, defines how much unbalanced the problem is
      on the second marginal.
    rank:
      Rank constraint on the coupling to minimize the linear OT problem
      :cite:`scetbon:21`. If :math:`-1`, no rank constraint is used.
    kwargs: Keyword arguments for
      :class:`~ott.solvers.linear.sinkhorn.Sinkhorn` or
      :class:`~ott.solvers.linear.sinkhorn_lr.LRSinkhorn`,
      depending on the ``rank``.

  Returns:
    The Sinkhorn output.
  """
  prob = linear_problem.LinearProblem(geom, a=a, b=b, tau_a=tau_a, tau_b=tau_b)
  if rank > 0:
    solver = sinkhorn_lr.LRSinkhorn(rank=rank, **kwargs)
  else:
    solver = sinkhorn.Sinkhorn(**kwargs)
  return solver(prob)


def solve_univariate(
    geom: pointcloud.PointCloud,
    a: Optional[jnp.ndarray] = None,
    b: Optional[jnp.ndarray] = None,
    *,
    return_transport: bool = False,
    return_dual_variables: bool = False,
) -> univariate.UnivariateOutput:
  """Solve 1D OT problems between two :math:`d`-dimensional point clouds.

  This function selects the underlying solver based on the following criteria:

  - :func:`~ott.solvers.linear.univariate.north_west_solver` - if
    ``return_dual_variables = True``.
  - :func:`~ott.solvers.linear.univariate.uniform_solver` - if ``a`` and
    ``b`` are both uniform and have the same size.
  - :func:`~ott.solvers.linear.univariate.quantile_solver` - otherwise.

  Args:
    geom: Geometry containing two :math:`d`-dimensional point clouds and
      a ground :class:`translation-invariant cost <ott.geometry.costs.TICost>`.
    a: The first marginal. If :obj:`None`, it will be uniform.
    b: The second marginal. If :obj:`None`, it will be uniform.
    return_transport: Whether to also return the mapped pairs used to compute
      the :attr:`~ott.solvers.linear.univariate.UnivariateOutput.transport_matrices`.
    return_dual_variables: Whether to also return the dual variables.

  Returns:
    The univariate output.
  """  # noqa: E501
  prob = linear_problem.LinearProblem(geom, a=a, b=b)
  if return_dual_variables:
    return univariate.north_west_solver(prob)
  if prob.is_uniform and prob.is_equal_size:
    return univariate.uniform_solver(prob, return_transport=return_transport)
  return univariate.quantile_solver(prob, return_transport=return_transport)



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import List, Optional, Sequence, Tuple, Union

import jax
import jax.numpy as jnp
import numpy as np
import scipy.sparse as sp

from ott.experimental import mmsinkhorn
from ott.geometry import pointcloud
from ott.solvers.linear import sinkhorn, sinkhorn_lr
from ott.solvers.quadratic import gromov_wasserstein

try:
  import matplotlib.colors as mcolors
  import matplotlib.patches as ptc
  import matplotlib.pyplot as plt
  from matplotlib import animation
except ImportError:
  plt = animation = None

# TODO(michalk8): make sure all outputs conform to a unified transport interface
Transport = Union[sinkhorn.SinkhornOutput, sinkhorn_lr.LRSinkhornOutput,
                  gromov_wasserstein.GWOutput]


@jax.jit
def ccworder(A: jnp.ndarray) -> jnp.ndarray:
  """Helper function to plot good-looking polygons.

  https://stackoverflow.com/questions/5040412/how-to-draw-the-largest-polygon-from-a-set-of-points
  """
  A = A - jnp.mean(A, 0, keepdims=True)
  return jnp.argsort(jnp.arctan2(A[:, 1], A[:, 0]))


def bidimensional(x: jnp.ndarray,
                  y: jnp.ndarray) -> Tuple[jnp.ndarray, jnp.ndarray]:
  """Apply PCA to reduce to bi-dimensional data."""
  if x.shape[1] < 3:
    return x, y

  u, s, _ = sp.linalg.svds(np.array(jnp.concatenate([x, y], axis=0)), k=2)
  proj = u * s
  k = x.shape[0]
  return proj[:k], proj[k:]


class Plot:
  """Plot an optimal transport map between two point clouds.

  This object can either plot or update a plot, to create animations as a
  :class:`~matplotlib.animation.FuncAnimation`, which can in turned be saved to
  disk at will. There are two design principles here:

  #. we do not rely on saving to/loading from disk to create animations
  #. we try as much as possible to disentangle the transport problem from
     its visualization.

  We use 2D scatter plots by default, relying on PCA visualization for d>3 data.
  This step requires a conversion to a numpy array, in order to compute leading
  singular values. This tool is therefore not designed having performance in
  mind.

  Args:
    fig: Specify figure object. Created by default
    ax: Specify axes objects. Created by default
    threshold: value below which links in transportation matrix won't be
      plotted. This value should be negative when using animations.
    scale: scale used for marker plots.
    show_lines: whether to show OT lines, as described in ``ot.matrix`` argument
    cmap: color map used to plot line colors.
    scale_alpha_by_coupling: use or not the coupling's value as proxy for alpha
    alpha: default alpha value for lines.
    title: title of the plot.
  """

  def __init__(
      self,
      fig: Optional["plt.Figure"] = None,
      ax: Optional["plt.Axes"] = None,
      threshold: float = -1.0,
      scale: int = 200,
      show_lines: bool = True,
      cmap: str = "cool",
      scale_alpha_by_coupling: bool = False,
      alpha: float = 0.7,
      title: Optional[str] = None
  ):
    if plt is None:
      raise RuntimeError("Please install `matplotlib` first.")

    if ax is None and fig is None:
      fig, ax = plt.subplots()
    elif fig is None:
      fig = plt.gcf()
    elif ax is None:
      ax = plt.gca()
    self.fig = fig
    self.ax = ax
    self._show_lines = show_lines
    self._lines = []
    self._points_x = None
    self._points_y = None
    self._threshold = threshold
    self._scale = scale
    self._cmap = cmap
    self._scale_alpha_by_coupling = scale_alpha_by_coupling
    self._alpha = alpha
    self._title = title

  def _scatter(self, ot: Transport):
    """Compute the position and scales of the points on a 2D plot."""
    if not isinstance(ot.geom, pointcloud.PointCloud):
      raise ValueError("So far we only plot PointCloud geometry.")

    x, y = ot.geom.x, ot.geom.y
    a, b = ot.a, ot.b
    x, y = bidimensional(x, y)
    scales_x = a * self._scale * a.shape[0]
    scales_y = b * self._scale * b.shape[0]
    return x, y, scales_x, scales_y

  def _mapping(self, x: jnp.ndarray, y: jnp.ndarray, matrix: jnp.ndarray):
    """Compute the lines representing the mapping between the 2 point clouds."""
    # Only plot the lines with a cost above the threshold.
    u, v = jnp.where(matrix > self._threshold)
    c = matrix[jnp.where(matrix > self._threshold)]
    xy = jnp.concatenate([x[u], y[v]], axis=-1)

    # Check if we want to adjust transparency.
    scale_alpha_by_coupling = self._scale_alpha_by_coupling

    # We can only adjust transparency if max(c) != min(c).
    if scale_alpha_by_coupling:
      min_matrix, max_matrix = jnp.min(c), jnp.max(c)
      scale_alpha_by_coupling = max_matrix != min_matrix

    result = []
    for i in range(xy.shape[0]):
      strength = jnp.max(jnp.array(matrix.shape)) * c[i]
      if scale_alpha_by_coupling:
        normalized_strength = (c[i] - min_matrix) / (max_matrix - min_matrix)
        alpha = self._alpha * float(normalized_strength)
      else:
        alpha = self._alpha

      # Matplotlib's transparency is sensitive to numerical errors.
      alpha = np.clip(alpha, 0.0, 1.0)

      start, end = xy[i, [0, 2]], xy[i, [1, 3]]
      result.append((start, end, strength, alpha))

    return result

  def __call__(self, ot: Transport) -> List["plt.Artist"]:
    """Plot couplings in 2-D, using PCA if data is higher dimensional."""
    x, y, sx, sy = self._scatter(ot)
    self._points_x = self.ax.scatter(
        *x.T, s=sx, edgecolors="k", marker="o", label="x"
    )
    self._points_y = self.ax.scatter(
        *y.T, s=sy, edgecolors="k", marker="X", label="y"
    )
    self.ax.legend(fontsize=15)
    if not self._show_lines:
      return []

    lines = self._mapping(x, y, ot.matrix)
    cmap = plt.get_cmap(self._cmap)
    self._lines = []
    for start, end, strength, alpha in lines:
      line, = self.ax.plot(
          start,
          end,
          linewidth=0.5 + 4 * strength,
          color=cmap(strength),
          zorder=0,
          alpha=alpha
      )
      self._lines.append(line)
    if self._title is not None:
      self.ax.set_title(self._title)
    return [self._points_x, self._points_y] + self._lines

  def update(self,
             ot: Transport,
             title: Optional[str] = None) -> List["plt.Artist"]:
    """Update a plot with a transport instance."""
    x, y, _, _ = self._scatter(ot)
    self._points_x.set_offsets(x)
    self._points_y.set_offsets(y)
    if not self._show_lines:
      return []

    new_lines = self._mapping(x, y, ot.matrix)
    cmap = plt.get_cmap(self._cmap)
    for line, new_line in zip(self._lines, new_lines):
      start, end, strength, alpha = new_line

      line.set_data(start, end)
      line.set_linewidth(0.5 + 4 * strength)
      line.set_color(cmap(strength))
      line.set_alpha(alpha)

    # Maybe add new lines to the plot.
    num_lines = len(self._lines)
    num_to_plot = len(new_lines) if self._show_lines else 0
    for i in range(num_lines, num_to_plot):
      start, end, strength, alpha = new_lines[i]

      line, = self.ax.plot(
          start,
          end,
          linewidth=0.5 + 4 * strength,
          color=cmap(strength),
          zorder=0,
          alpha=alpha
      )
      self._lines.append(line)

    self._lines = self._lines[:num_to_plot]  # Maybe remove some
    if title is not None:
      self.ax.set_title(title)
    return [self._points_x, self._points_y] + self._lines

  def animate(
      self,
      transports: Sequence[Transport],
      titles: Optional[Sequence[str]] = None,
      frame_rate: float = 10.0
  ) -> "animation.FuncAnimation":
    """Make an animation from several transports."""
    _ = self(transports[0])
    if titles is None:
      titles = [None for _ in np.arange(0, len(transports))]
    assert len(titles) == len(transports), (
        f"titles/transports lengths differ `{len(titles)}`/`{len(transports)}`."
    )
    return animation.FuncAnimation(
        self.fig,
        lambda i: self.update(transports[i], titles[i]),
        np.arange(0, len(transports)),
        init_func=lambda: self.update(transports[0], titles[0]),
        interval=1000 / frame_rate,
        blit=True
    )


# TODO(zoepiran): add support for data of d > 2 (PCA on all k's)
class PlotMM(Plot):
  """Plots an optimal transport map for :class:`~ott.experimental.mmsinkhorn.MMSinkhorn`.

  It enables to either plot or update a plot in a single object, offering the
  possibilities to create animations as a
  :class:`~matplotlib.animation.FuncAnimation`, which can in turned be saved to
  disk at will. There are two design principles here:

  #. we do not rely on saving to/loading from disk to create animations
  #. we try as much as possible to disentangle the transport problem from
       its visualization.

  Args:
    fig: Specify figure object. Created by default
    ax: Specify axes objects. Created by default
    fix_axes_lim: Whether to fix x/y limits to :math:`[0, 1]`.
    cmap: color map used to plot line colors.
    markers: Markers for each marginal.
    alpha: default alpha value for lines.
    title: title of the plot.
  """  # noqa: E501

  def __init__(
      self,
      fig: Optional["plt.Figure"] = None,
      ax: Optional["plt.Axes"] = None,
      fix_axes_lim: bool = False,
      cmap: Union[str, "mcolors.Colormap"] = "cividis_r",
      markers: str = "svopxdh",
      alpha: float = 0.6,
      title: Optional[str] = None,
  ):
    if plt is None:
      raise RuntimeError("Please install `matplotlib` first.")
    if isinstance(cmap, str):
      cmap = plt.colormaps[cmap]
    super().__init__(fig=fig, ax=ax, cmap=cmap, alpha=alpha, title=title)
    self._patches = None
    self._points = None
    self._markers = markers
    self._fix_axes_lim = fix_axes_lim

  def __call__(
      self,
      ot: mmsinkhorn.MMSinkhornOutput,
      top_k: Optional[int] = None
  ) -> List["plt.Artist"]:
    """Plot 2-D couplings. does not support higher dimensional."""
    assert ot.n_marginals <= len(self._markers), "Not enough markers to plot."
    self._points = []
    self._patches = []
    n0 = max(ot.shape)
    top_k = n0 if top_k is None else top_k

    # extract the `top_k` entries in the tensor, and their indices.
    _, idx = jax.lax.top_k(ot.tensor.ravel(), top_k)
    indices = jnp.unravel_index(idx, ot.shape)

    alphas = np.linspace(self._alpha, 0.2, max(0, top_k - n0))
    for j in range(top_k):
      points = [x[indices[i][j], ...] for i, x in enumerate(ot.x_s)]
      # re-order to ensure polygons have maximal area
      points = [points[i] for i in ccworder(jnp.array(points))]
      alpha = self._alpha if j < n0 else alphas[j - n0]

      polygon = ptc.Polygon(
          points,
          fill=True,
          linewidth=2,
          color=self._cmap(float(j >= n0)),
          alpha=alpha,
          zorder=-j,
      )
      self._patches.append(self.ax.add_patch(polygon))

    for i, (x, a) in enumerate(zip(ot.x_s, ot.a_s)):
      points = self.ax.scatter(
          x[:, 0],
          x[:, 1],
          s=200.0 * len(a) * a,
          marker=self._markers[i],
          c="black",
          linewidth=0.0,
          edgecolor=None,
          label=str(i)
      )
      self._points.append(points)

    if self._title is not None:
      self.ax.set_title(self._title)

    return self._points + self._patches

  def update(
      self,
      ot: mmsinkhorn.MMSinkhornOutput,
      title: Optional[str] = None,
      top_k: Optional[int] = None,
  ) -> List["plt.Artist"]:
    """Update a plot with a transport instance."""
    n0 = max(ot.shape)
    top_k = n0 if top_k is None else top_k
    # extract the `top_k` entries in the tensor, and their indices.
    _, idx = jax.lax.top_k(ot.tensor.ravel(), top_k)
    indices = jnp.unravel_index(idx, ot.shape)

    alphas = np.linspace(self._alpha, 0.2, max(0, top_k - n0))
    for j, patch in enumerate(self._patches):
      points = [x[indices[i][j], ...] for i, x in enumerate(ot.x_s)]
      # re-order to ensure polygons have maximal area
      points = [points[i] for i in ccworder(jnp.array(points))]
      alpha = self._alpha if j < n0 else alphas[j - n0]
      # update the location of the patches according to the new coordinates
      patch.set_xy(points)
      patch.set_color(self._cmap(float(j >= n0)))
      patch.set_alpha(alpha)

    for points, xs in zip(self._points, ot.x_s):
      points.set_offsets(xs)

    if title is not None:
      self.ax.set_title(title)

    # we keep the axis fixed to 0-1 assuming normalized data
    if self._fix_axes_lim:
      eps = 2.5e-2
      self.ax.set_ylim(-eps, 1.0 + eps)
      self.ax.set_xlim(-eps, 1.0 + eps)

    return self._points + self._patches

  def animate(
      self,
      transports: Sequence[mmsinkhorn.MMSinkhornOutput],
      titles: Optional[Sequence[str]] = None,
      frame_rate: float = 10.0,
      top_k: Optional[int] = None,
  ) -> "animation.FuncAnimation":
    """Make an animation from several transports."""
    ot, *_ = transports
    _ = self(ot, top_k=top_k)
    titles = titles if titles is not None else [""] * len(transports)
    return animation.FuncAnimation(
        self.fig,
        lambda i: self.update(ot=transports[i], title=titles[i], top_k=top_k),
        np.arange(0, len(transports)),
        init_func=lambda: self.update(ot, title=titles[0], top_k=top_k),
        interval=1000.0 / frame_rate,
        blit=True,
    )



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple

import jax
import jax.numpy as jnp
import jax.tree_util as jtu
import numpy as np

import diffrax
from flax.training import train_state

from ott import utils
from ott.neural.methods.flows import dynamics
from ott.neural.networks import velocity_field
from ott.solvers import utils as solver_utils

__all__ = ["OTFlowMatching"]


class OTFlowMatching:
  """(Optimal transport) flow matching :cite:`lipman:22`.

  With an extension to OT-FM :cite:`tong:23,pooladian:23`.

  Args:
    vf: Vector field parameterized by a neural network.
    flow: Flow between the source and the target distributions.
    match_fn: Function to match samples from the source and the target
      distributions. It has a ``(src, tgt) -> matching`` signature.
    time_sampler: Time sampler with a ``(rng, n_samples) -> time`` signature.
    kwargs: Keyword arguments for
      :meth:`~ott.neural.networks.velocity_field.VelocityField.create_train_state`.
  """

  def __init__(
      self,
      vf: velocity_field.VelocityField,
      flow: dynamics.BaseFlow,
      match_fn: Optional[Callable[[jnp.ndarray, jnp.ndarray],
                                  jnp.ndarray]] = None,
      time_sampler: Callable[[jax.Array, int],
                             jnp.ndarray] = solver_utils.uniform_sampler,
      **kwargs: Any,
  ):
    self.vf = vf
    self.flow = flow
    self.time_sampler = time_sampler
    self.match_fn = match_fn

    self.vf_state = self.vf.create_train_state(
        input_dim=self.vf.output_dims[-1], **kwargs
    )
    self.step_fn = self._get_step_fn()

  def _get_step_fn(self) -> Callable:

    @jax.jit
    def step_fn(
        rng: jax.Array,
        vf_state: train_state.TrainState,
        source: jnp.ndarray,
        target: jnp.ndarray,
        source_conditions: Optional[jnp.ndarray],
    ) -> Tuple[Any, Any]:

      def loss_fn(
          params: jnp.ndarray, t: jnp.ndarray, source: jnp.ndarray,
          target: jnp.ndarray, source_conditions: Optional[jnp.ndarray],
          rng: jax.Array
      ) -> jnp.ndarray:
        rng_flow, rng_dropout = jax.random.split(rng, 2)
        x_t = self.flow.compute_xt(rng_flow, t, source, target)
        v_t = vf_state.apply_fn({"params": params},
                                t,
                                x_t,
                                source_conditions,
                                rngs={"dropout": rng_dropout})
        u_t = self.flow.compute_ut(t, source, target)

        return jnp.mean((v_t - u_t) ** 2)

      batch_size = len(source)
      key_t, key_model = jax.random.split(rng, 2)
      t = self.time_sampler(key_t, batch_size)
      grad_fn = jax.value_and_grad(loss_fn)
      loss, grads = grad_fn(
          vf_state.params, t, source, target, source_conditions, key_model
      )
      return vf_state.apply_gradients(grads=grads), loss

    return step_fn

  def __call__(  # noqa: D102
      self,
      loader: Iterable[Dict[str, np.ndarray]],
      *,
      n_iters: int,
      rng: Optional[jax.Array] = None,
  ) -> Dict[str, List[float]]:
    """Train the OTFlowMatching model.

    Args:
      loader: Data loader returning a dictionary with possible keys
      `src_lin`, `tgt_lin`, `src_condition`.
      n_iters: Number of iterations to train the model.
      rng: Random number generator.

    Returns:
      Training logs.
    """
    rng = utils.default_prng_key(rng)
    training_logs = {"loss": []}
    for batch in loader:
      rng, rng_resample, rng_step_fn = jax.random.split(rng, 3)

      batch = jtu.tree_map(jnp.asarray, batch)

      src, tgt = batch["src_lin"], batch["tgt_lin"]
      src_cond = batch.get("src_condition")

      if self.match_fn is not None:
        tmat = self.match_fn(src, tgt)
        src_ixs, tgt_ixs = solver_utils.sample_joint(rng_resample, tmat)
        src, tgt = src[src_ixs], tgt[tgt_ixs]
        src_cond = None if src_cond is None else src_cond[src_ixs]

      self.vf_state, loss = self.step_fn(
          rng_step_fn,
          self.vf_state,
          src,
          tgt,
          src_cond,
      )

      training_logs["loss"].append(float(loss))
      if len(training_logs["loss"]) >= n_iters:
        break

    return training_logs

  def transport(
      self,
      x: jnp.ndarray,
      condition: Optional[jnp.ndarray] = None,
      t0: float = 0.0,
      t1: float = 1.0,
      **kwargs: Any,
  ) -> jnp.ndarray:
    """Transport data with the learned map.

    This method pushes-forward the data by solving the neural ODE
    parameterized by the velocity field.

    Args:
      x: Initial condition of the ODE of shape ``[batch_size, ...]``.
      condition: Condition of the input data of shape ``[batch_size, ...]``.
      t0: Starting point of integration.
      t1: End point of integration.
      kwargs: Keyword arguments for the ODE solver.

    Returns:
      The push-forward or pull-back distribution defined by the learned
      transport plan.
    """

    def vf(
        t: jnp.ndarray, x: jnp.ndarray, cond: Optional[jnp.ndarray]
    ) -> jnp.ndarray:
      params = self.vf_state.params
      return self.vf_state.apply_fn({"params": params}, t, x, cond, train=False)

    def solve_ode(x: jnp.ndarray, cond: Optional[jnp.ndarray]) -> jnp.ndarray:
      ode_term = diffrax.ODETerm(vf)
      result = diffrax.diffeqsolve(
          ode_term,
          t0=t0,
          t1=t1,
          y0=x,
          args=cond,
          **kwargs,
      )
      return result.ys[0]

    kwargs.setdefault("dt0", None)
    kwargs.setdefault("solver", diffrax.Tsit5())
    kwargs.setdefault(
        "stepsize_controller", diffrax.PIDController(rtol=1e-5, atol=1e-5)
    )

    in_axes = [0, None if condition is None else 0]
    return jax.jit(jax.vmap(solve_ode, in_axes))(x, condition)



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Callable, Optional, Sequence

import jax
import jax.numpy as jnp

import optax
from flax import linen as nn
from flax.training import train_state

from ott.neural.networks.layers import time_encoder

__all__ = ["VelocityField"]


class VelocityField(nn.Module):
  r"""Neural vector field.

  This class learns a map :math:`v: \mathbb{R}\times \mathbb{R}^d
  \rightarrow \mathbb{R}^d` solving the ODE :math:`\frac{dx}{dt} = v(t, x)`.
  Given a source distribution at time :math:`t_0`, the velocity field can be
  used to transport the source distribution given at :math:`t_0` to
  a target distribution given at :math:`t_1` by integrating :math:`v(t, x)`
  from :math:`t=t_0` to :math:`t=t_1`.

  Args:
    hidden_dims: Dimensionality of the embedding of the data.
    output_dims: Dimensionality of the embedding of the output.
    condition_dims: Dimensionality of the embedding of the condition.
      If :obj:`None`, the velocity field has no conditions.
    time_dims: Dimensionality of the time embedding.
      If :obj:`None`, ``hidden_dims`` is used.
    time_encoder: Time encoder for the velocity field.
    act_fn: Activation function.
  """
  hidden_dims: Sequence[int]
  output_dims: Sequence[int]
  condition_dims: Optional[Sequence[int]] = None
  time_dims: Optional[Sequence[int]] = None
  time_encoder: Callable[[jnp.ndarray],
                         jnp.ndarray] = time_encoder.cyclical_time_encoder
  act_fn: Callable[[jnp.ndarray], jnp.ndarray] = nn.silu
  dropout_rate: float = 0.0

  @nn.compact
  def __call__(
      self,
      t: jnp.ndarray,
      x: jnp.ndarray,
      condition: Optional[jnp.ndarray] = None,
      train: bool = True,
  ) -> jnp.ndarray:
    """Forward pass through the neural vector field.

    Args:
      t: Time of shape ``[batch, 1]``.
      x: Data of shape ``[batch, ...]``.
      condition: Conditioning vector of shape ``[batch, ...]``.
      train: If `True`, enables dropout for training.

    Returns:
      Output of the neural vector field of shape ``[batch, output_dim]``.
    """
    time_dims = self.hidden_dims if self.time_dims is None else self.time_dims

    t = self.time_encoder(t)
    for time_dim in time_dims:
      t = self.act_fn(nn.Dense(time_dim)(t))
      t = nn.Dropout(rate=self.dropout_rate, deterministic=not train)(t)

    for hidden_dim in self.hidden_dims:
      x = self.act_fn(nn.Dense(hidden_dim)(x))
      x = nn.Dropout(rate=self.dropout_rate, deterministic=not train)(x)

    if self.condition_dims is not None:
      assert condition is not None, "No condition was passed."
      for cond_dim in self.condition_dims:
        condition = self.act_fn(nn.Dense(cond_dim)(condition))
        condition = nn.Dropout(
            rate=self.dropout_rate, deterministic=not train
        )(
            condition
        )
      feats = jnp.concatenate([t, x, condition], axis=-1)
    else:
      feats = jnp.concatenate([t, x], axis=-1)

    for output_dim in self.output_dims[:-1]:
      feats = self.act_fn(nn.Dense(output_dim)(feats))
      feats = nn.Dropout(rate=self.dropout_rate, deterministic=not train)(feats)

    # No activation function for the final layer
    return nn.Dense(self.output_dims[-1])(feats)

  def create_train_state(
      self,
      rng: jax.Array,
      optimizer: optax.OptState,
      input_dim: int,
      condition_dim: Optional[int] = None,
  ) -> train_state.TrainState:
    """Create the training state.

    Args:
      rng: Random number generator.
      optimizer: Optimizer.
      input_dim: Dimensionality of the velocity field.
      condition_dim: Dimensionality of the condition of the velocity field.

    Returns:
      The training state.
    """
    t, x = jnp.ones((1, 1)), jnp.ones((1, input_dim))
    if self.condition_dims is None:
      cond = None
    else:
      assert condition_dim > 0, "Condition dimension must be positive."
      cond = jnp.ones((1, condition_dim))

    params = self.init(rng, t, x, cond, train=False)["params"]
    return train_state.TrainState.create(
        apply_fn=self.apply, params=params, tx=optimizer
    )



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import (
    Any,
    Callable,
    Dict,
    Iterator,
    List,
    Literal,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import jax
import jax.numpy as jnp

import optax
from flax import linen as nn
from flax.core import frozen_dict

from ott import utils
from ott.geometry import costs
from ott.neural.networks import potentials
from ott.problems.linear import potentials as dual_potentials

__all__ = ["ENOTPotentials", "PotentialModelWrapper", "ExpectileNeuralDual"]

Train_t = Dict[Literal["train_logs", "valid_logs"], Dict[str, List[float]]]
Callback_t = Callable[[int, dual_potentials.DualPotentials], None]


@jax.tree_util.register_pytree_node_class
class ENOTPotentials(dual_potentials.DualPotentials):
  """The dual potentials of the ENOT method :cite:`buzun:24`.

  Args:
    grad_f: Gradient of the first dual potential function.
    g: The second dual potential function.
    cost_fn: The cost function used to solve the OT problem.
    is_bidirectional: Whether the duals are trained for bidirectional
      transport mapping.
    corr: Whether the duals solve the problem in correlation form.
  """

  def __init__(
      self, grad_f: potentials.PotentialGradientFn_t,
      g: potentials.PotentialValueFn_t, cost_fn: costs.CostFn, *,
      is_bidirectional: bool, corr: bool
  ):
    self.__grad_f = grad_f
    self.is_bidirectional = is_bidirectional

    def g_cost_conjugate(x: jnp.ndarray) -> jnp.ndarray:
      if is_bidirectional and not corr:
        y_hat = cost_fn.twist_operator(x, grad_f(x), False)
      else:
        y_hat = grad_f(x)
      y_hat = jax.lax.stop_gradient(y_hat)

      return -g(y_hat) + (jnp.dot(x, y_hat) if corr else cost_fn(x, y_hat))

    super().__init__(g_cost_conjugate, g, cost_fn=cost_fn, corr=corr)

  @property
  def _grad_f(self) -> Callable[[jnp.ndarray], jnp.ndarray]:
    return jax.vmap(self.__grad_f)

  def transport(  # noqa: D102
      self,
      vec: jnp.ndarray,
      forward: bool = True
  ) -> jnp.ndarray:
    if self.is_bidirectional:
      return super().transport(vec, forward)
    vec = jnp.atleast_2d(vec)
    assert forward, "Only forward mapping (source -> target) is supported."
    return self._grad_f(vec)

  def tree_flatten(self) -> Tuple[Sequence[Any], Dict[str, Any]]:
    """Flatten the kwargs."""
    return [], {
        "grad_f": self.__grad_f,
        "g": self._g,
        "cost_fn": self.cost_fn,
        "is_bidirectional": self.is_bidirectional,
        "corr": self._corr
    }


class PotentialModelWrapper(potentials.BasePotential):
  """Wrapper class for the neural models.

  Implements a potential value or a vector field.

  Args:
    model: Network architecture of the potential.
    add_l2_norm: If :obj:`True`, l2 norm is added to the potential.
    is_potential: Model the potential if ``True``, otherwise
      model the gradient of the potential.
  """

  model: nn.Module
  add_l2_norm: bool
  is_potential: bool = True

  @nn.compact
  def __call__(self, x: jnp.ndarray) -> jnp.ndarray:
    """Apply model and optionally add l2 norm or x."""
    z: jnp.ndarray = self.model(x)

    if self.is_potential:
      z = z.squeeze()

    if self.add_l2_norm:
      z = z + (0.5 * jnp.dot(x, x)) if self.is_potential else x

    return z

  def potential_gradient_fn(
      self, params: frozen_dict.FrozenDict[str, jnp.ndarray]
  ) -> potentials.PotentialGradientFn_t:
    """A vector function or gradient of the potential."""
    if self.is_potential:
      return jax.grad(self.potential_value_fn(params))
    return lambda x: self.apply({"params": params}, x)


class ExpectileNeuralDual:
  r"""Expectile-regularized Neural Optimal Transport (ENOT) :cite:`buzun:24`.

  It solves the dual optimal transport problem for a specified cost function
  :math:`c(x, y)` between two measures :math:`\alpha` and :math:`\beta` in
  :math:`d`-dimensional Euclidean space with additional regularization on
  Kantorovich potentials. The expectile regularization enforces binding
  conditions on the learning dual potentials :math:`f` and :math:`g`.
  The main optimization objective is

  .. math::

    \sup_{g \in L_1(\beta)} \inf_{T: \, R^d \to R^d} \big[
      \mathbb{E}_{\alpha}[c(x, T(x))] + \mathbb{E}_{\beta} [g(y)]
      - \mathbb{E}_{\alpha} [g(T(x))]
    \big],

  where :math:`T(x)` is the transport mapping from :math:`\alpha`
  to :math:`\beta` expressed through :math:`\nabla f(x)`. The explicit
  formula depends on the cost function and ``is_bidirectional`` training
  option. The regularization term is

  .. math::

    \mathbb{E} \mathcal{L}_{\tau} \big(
      c(x, T(x)) - g(T(x)) - c(x, y)  + g(y)
    \big),

  where :math:`\mathcal{L}_{\tau}` is the least asymmetrically weighted
  squares loss from expectile regression.

  The potentials for ``neural_f`` and ``neural_g`` can

  1. both provide the values of the potentials :math:`f` and :math:`g`, or
  2. when parameter ``is_bidirectional=False``, ``neural_f`` provides
     the gradient :math:`\nabla f` for mapping :math:`T`.

  Args:
    dim_data: Input dimensionality of data required for network init.
    neural_f: Network architecture for potential :math:f or
      its gradient :math:`\nabla f`.
    neural_g: Network architecture for potential :math:`g`.
    optimizer_f: Optimizer function for potential :math:`f`.
    optimizer_g: Optimizer function for potential :math:`g`.
    cost_fn: Cost function of the OT problem.
    is_bidirectional: Alternate between updating the forward and backward
      directions. Inspired from :cite:`jacobs:20`.
    use_dot_product: Whether the duals solve the problem in correlation form.
    expectile: Parameter of the expectile loss (:math:`\tau`).
      Suggested values range is :math:`[0.9, 1.0)`.
    expectile_loss_coef: Expectile loss weight.
      Suggested values range is :math:`[0.3, 1.0]`.
    num_train_iters: Number of total training iterations.
    valid_freq: Frequency with which model is validated.
    log_freq: Frequency with training and validation are logged.
    logging: Option to return logs.
    rng: Random key used for seeding for network initializations.
  """

  def __init__(
      self,
      dim_data: int,
      neural_f: Optional[nn.Module] = None,
      neural_g: Optional[nn.Module] = None,
      optimizer_f: Optional[optax.GradientTransformation] = None,
      optimizer_g: Optional[optax.GradientTransformation] = None,
      cost_fn: Optional[costs.CostFn] = None,
      is_bidirectional: bool = True,
      use_dot_product: bool = False,
      expectile: float = 0.99,
      expectile_loss_coef: float = 1.0,
      num_train_iters: int = 20000,
      valid_freq: int = 1000,
      log_freq: int = 1000,
      logging: bool = False,
      rng: Optional[jax.Array] = None
  ):
    self.num_train_iters = num_train_iters
    self.valid_freq = valid_freq
    self.log_freq = log_freq
    self.logging = logging
    self.cost_fn = costs.SqEuclidean() if cost_fn is None else cost_fn
    self.expectile = expectile
    self.expectile_loss_coef = expectile_loss_coef
    self.is_bidirectional = is_bidirectional
    self.use_dot_product = use_dot_product

    if is_bidirectional:
      assert isinstance(self.cost_fn, costs.TICost), (
          "is_bidirectional=True can only be used with a translation invariant"
          "cost (TICost)"
      )

    if use_dot_product:
      assert isinstance(
          self.cost_fn, costs.SqEuclidean
      ), ("use_dot_product=True can only be used with SqEuclidean cost")

    if use_dot_product:
      self.train_batch_cost = lambda x, y: -jax.vmap(jnp.dot)(x, y)
    else:
      self.train_batch_cost = jax.vmap(self.cost_fn)

    # set default optimizers
    if optimizer_f is None:
      optimizer_f = optax.adam(learning_rate=3e-4, b1=0.9, b2=0.9)
    if optimizer_g is None:
      optimizer_g = optax.adam(learning_rate=3e-4, b1=0.9, b2=0.9)

    # set default neural architectures
    if neural_f is None:
      last_dim = 1 if is_bidirectional else dim_data
      neural_f = potentials.MLP(
          dim_hidden=[128, 128, 128, 128, last_dim], act_fn=jax.nn.elu
      )
    if neural_g is None:
      neural_g = potentials.MLP(
          dim_hidden=[128, 128, 128, 128, 1], act_fn=jax.nn.elu
      )

    self.neural_f = PotentialModelWrapper(
        model=neural_f,
        is_potential=is_bidirectional,
        add_l2_norm=self.use_dot_product
    )
    self.neural_g = PotentialModelWrapper(
        model=neural_g, is_potential=True, add_l2_norm=self.use_dot_product
    )

    rng = utils.default_prng_key(rng)
    rng_f, rng_g = jax.random.split(rng, 2)

    self.state_f = self.neural_f.create_train_state(
        rng_f, optimizer_f, (dim_data,)
    )
    self.state_g = self.neural_g.create_train_state(
        rng_g, optimizer_g, (dim_data,)
    )

    self.train_step = self._get_train_step()
    self.valid_step = self._get_valid_step()

  def __call__(
      self,
      trainloader_source: Iterator[jnp.ndarray],
      trainloader_target: Iterator[jnp.ndarray],
      validloader_source: Iterator[jnp.ndarray],
      validloader_target: Iterator[jnp.ndarray],
      callback: Optional[Callback_t] = None,
  ) -> Union[ENOTPotentials, Tuple[ENOTPotentials, Train_t]]:
    """Train and return the Kantorovich dual potentials."""
    logs = self.train_fn(
        trainloader_source,
        trainloader_target,
        validloader_source,
        validloader_target,
        callback=callback,
    )
    res = self.to_dual_potentials()

    return (res, logs) if self.logging else res

  def train_fn(
      self,
      trainloader_source: Iterator[jnp.ndarray],
      trainloader_target: Iterator[jnp.ndarray],
      validloader_source: Iterator[jnp.ndarray],
      validloader_target: Iterator[jnp.ndarray],
      callback: Optional[Callback_t] = None,
  ) -> Train_t:
    """Training and validation."""
    try:
      from tqdm.auto import tqdm
    except ImportError:
      tqdm = lambda _: _

    train_batch, valid_batch = {}, {}

    train_logs = {"loss_f": [], "loss_g": [], "w_dist": [], "directions": []}
    valid_logs = {"loss_f": [], "loss_g": [], "w_dist": []}

    for step in tqdm(range(self.num_train_iters)):

      update_forward = (step % 2 == 0) or not self.is_bidirectional

      if update_forward:
        train_batch["source"] = jnp.asarray(next(trainloader_source))
        train_batch["target"] = jnp.asarray(next(trainloader_target))
        (self.state_f, self.state_g, loss, loss_f, loss_g,
         w_dist) = self.train_step(self.state_f, self.state_g, train_batch)
      else:
        train_batch["target"] = jnp.asarray(next(trainloader_source))
        train_batch["source"] = jnp.asarray(next(trainloader_target))
        (self.state_g, self.state_f, loss, loss_g, loss_f,
         w_dist) = self.train_step(self.state_g, self.state_f, train_batch)

      if self.logging and step % self.log_freq == 0:
        self._update_logs(train_logs, loss_f, loss_g, w_dist)

      if callback is not None:
        _ = callback(step, self.to_dual_potentials())

      if step != 0 and step % self.valid_freq == 0:
        valid_batch["source"] = jnp.asarray(next(validloader_source))
        valid_batch["target"] = jnp.asarray(next(validloader_target))

        valid_loss_f, valid_loss_g, valid_w_dist = self.valid_step(
            self.state_f, self.state_g, valid_batch
        )

        if self.logging:
          self._update_logs(
              valid_logs, valid_loss_f, valid_loss_g, valid_w_dist
          )

    return {"train_logs": train_logs, "valid_logs": valid_logs}

  def _get_train_step(
      self
  ) -> Callable[[
      potentials.PotentialTrainState, potentials.PotentialTrainState, Dict[
          str, jnp.ndarray]
  ], Tuple[potentials.PotentialTrainState, potentials.PotentialTrainState,
           jnp.ndarray, jnp.ndarray, jnp.ndarray, jnp.ndarray]]:

    @jax.jit
    def step_fn(state_f, state_g, batch):
      grad_fn = jax.value_and_grad(self._loss_fn, argnums=[0, 1], has_aux=True)
      (loss, (loss_f, loss_g, w_dist)), (grads_f, grads_g) = grad_fn(
          state_f.params,
          state_g.params,
          state_f.potential_gradient_fn,
          state_g.potential_value_fn,
          batch,
      )

      return (
          state_f.apply_gradients(grads=grads_f),
          state_g.apply_gradients(grads=grads_g), loss, loss_f, loss_g, w_dist
      )

    return step_fn

  def _get_valid_step(
      self
  ) -> Callable[[
      potentials.PotentialTrainState, potentials.PotentialTrainState, Dict[
          str, jnp.ndarray]
  ], Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray]]:

    @jax.jit
    def step_fn(state_f, state_g, batch):
      loss, (loss_f, loss_g, w_dist) = self._loss_fn(
          state_f.params,
          state_g.params,
          state_f.potential_gradient_fn,
          state_g.potential_value_fn,
          batch,
      )

      return loss_f, loss_g, w_dist

    return step_fn

  def _expectile_loss(self, diff: jnp.ndarray) -> jnp.ndarray:
    """Loss of the expectile regression :cite:`buzun:24`."""
    weight = jnp.where(diff >= 0, self.expectile, (1 - self.expectile))
    return weight * diff ** 2

  def _get_g_value_partial(
      self, params_g: frozen_dict.FrozenDict[str, jnp.ndarray],
      g_value: Callable[[frozen_dict.FrozenDict[str, jnp.ndarray]],
                        potentials.PotentialValueFn_t]
  ):

    if self.use_dot_product:
      g_value_partial = lambda y: -jax.vmap(g_value(params_g))(y)
      g_value_partial_detach = \
          lambda y: -jax.vmap(g_value(jax.lax.stop_gradient(params_g)))(y)
    else:
      g_value_partial = jax.vmap(g_value(params_g))
      g_value_partial_detach = jax.vmap(
          g_value(jax.lax.stop_gradient(params_g))
      )

    return g_value_partial, g_value_partial_detach

  def _distance(
      self, source: jnp.ndarray, target: jnp.ndarray, f_source: jnp.ndarray,
      g_target: jnp.ndarray
  ) -> jnp.ndarray:

    w_dist = f_source.mean() + g_target.mean()

    if self.use_dot_product:
      w_dist = jnp.mean(jnp.sum(source ** 2, axis=-1)) + \
               jnp.mean(jnp.sum(target ** 2, axis=-1)) + \
               2 * w_dist

    return w_dist

  def _loss_fn(
      self, params_f: frozen_dict.FrozenDict[str, jnp.ndarray],
      params_g: frozen_dict.FrozenDict[str, jnp.ndarray],
      gradient_f: Callable[[frozen_dict.FrozenDict[str, jnp.ndarray]],
                           potentials.PotentialGradientFn_t],
      g_value: Callable[[frozen_dict.FrozenDict[str, jnp.ndarray]],
                        potentials.PotentialValueFn_t], batch: Dict[str,
                                                                    jnp.ndarray]
  ) -> Tuple[jnp.ndarray, Tuple[jnp.ndarray, jnp.ndarray, jnp.ndarray]]:

    source, target = batch["source"], batch["target"]

    g_value_partial, g_value_partial_detach = self._get_g_value_partial(
        params_g, g_value
    )
    batch_cost = self.train_batch_cost

    transport = ENOTPotentials(
        gradient_f(params_f),
        g_value(params_g),
        self.cost_fn,
        is_bidirectional=self.is_bidirectional,
        corr=self.use_dot_product
    ).transport

    target_hat = transport(source)
    target_hat_detach = jax.lax.stop_gradient(target_hat)

    g_target = g_value_partial(target)
    g_star_source = batch_cost(source, target_hat_detach)\
      - g_value_partial(target_hat_detach)

    diff_1 = jax.lax.stop_gradient(g_star_source - batch_cost(source, target))\
      + g_target
    reg_loss_1 = self._expectile_loss(diff_1).mean()

    diff_2 = jax.lax.stop_gradient(g_target - batch_cost(source, target))\
      + g_star_source
    reg_loss_2 = self._expectile_loss(diff_2).mean()

    reg_loss = (reg_loss_1 + reg_loss_2) * self.expectile_loss_coef
    dual_loss = -(g_target + g_star_source).mean()
    amor_loss = (
        batch_cost(source, target_hat) - g_value_partial_detach(target_hat)
    ).mean()

    loss = reg_loss + dual_loss + amor_loss
    f_loss = amor_loss
    g_loss = reg_loss + dual_loss
    w_dist = self._distance(source, target, g_star_source, g_target)

    return loss, (f_loss, g_loss, w_dist)

  def to_dual_potentials(self) -> ENOTPotentials:
    """Return the Kantorovich dual potentials from the trained potentials."""
    f_grad_partial = self.state_f.potential_gradient_fn(self.state_f.params)
    g_value_partial = self.state_g.potential_value_fn(self.state_g.params, None)

    return ENOTPotentials(
        f_grad_partial,
        g_value_partial,
        self.cost_fn,
        is_bidirectional=self.is_bidirectional,
        corr=self.use_dot_product
    )

  @staticmethod
  def _update_logs(
      logs: Dict[str, List[Union[float, str]]],
      loss_f: jnp.ndarray,
      loss_g: jnp.ndarray,
      w_dist: jnp.ndarray,
  ) -> None:
    logs["loss_f"].append(float(loss_f))
    logs["loss_g"].append(float(loss_g))
    logs["w_dist"].append(float(w_dist))



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import collections
import dataclasses
from typing import Any, Dict, Optional, Sequence

import numpy as np

__all__ = ["OTData", "OTDataset"]

Item_t = Dict[str, np.ndarray]


@dataclasses.dataclass(repr=False, frozen=True)
class OTData:
  """Distribution data for (conditional) optimal transport problems.

  Args:
    lin: Linear term of the samples.
    quad: Quadratic term of the samples.
    condition: Condition corresponding to the data distribution.
  """
  lin: Optional[np.ndarray] = None
  quad: Optional[np.ndarray] = None
  condition: Optional[np.ndarray] = None

  def __getitem__(self, ix: int) -> Item_t:
    return {k: v[ix] for k, v in self.__dict__.items() if v is not None}

  def __len__(self) -> int:
    if self.lin is not None:
      return len(self.lin)
    if self.quad is not None:
      return len(self.quad)
    return 0


class OTDataset:
  """Dataset for optimal transport problems.

  Args:
    src_data: Samples from the source distribution.
    tgt_data: Samples from the target distribution.
    src_conditions: Conditions for the source data.
    tgt_conditions: Conditions for the target data.
    is_aligned: Whether the samples from the source and the target data
      are paired. If yes, the source and the target conditions must match.
    seed: Random seed used to match source and target when not aligned.
  """
  SRC_PREFIX = "src"
  TGT_PREFIX = "tgt"

  def __init__(
      self,
      src_data: OTData,
      tgt_data: OTData,
      src_conditions: Optional[Sequence[Any]] = None,
      tgt_conditions: Optional[Sequence[Any]] = None,
      is_aligned: bool = False,
      seed: Optional[int] = None,
  ):
    self.src_data = src_data
    self.tgt_data = tgt_data

    if src_conditions is None:
      src_conditions = [None] * len(src_data)
    self.src_conditions = list(src_conditions)
    if tgt_conditions is None:
      tgt_conditions = [None] * len(tgt_data)
    self.tgt_conditions = list(tgt_conditions)

    self._tgt_cond_to_ix = collections.defaultdict(list)
    for ix, cond in enumerate(tgt_conditions):
      self._tgt_cond_to_ix[cond].append(ix)

    self.is_aligned = is_aligned
    self._rng = np.random.default_rng(seed)

    self._verify_integrity()

  def _verify_integrity(self) -> None:
    assert len(self.src_data) == len(self.src_conditions)
    assert len(self.tgt_data) == len(self.tgt_conditions)

    if self.is_aligned:
      assert len(self.src_data) == len(self.tgt_data)
      assert self.src_conditions == self.tgt_conditions
    else:
      sym_diff = set(self.src_conditions
                    ).symmetric_difference(self.tgt_conditions)
      assert not sym_diff, sym_diff

  def _sample_from_target(self, src_ix: int) -> Item_t:
    src_cond = self.src_conditions[src_ix]
    tgt_ixs = self._tgt_cond_to_ix[src_cond]
    ix = self._rng.choice(tgt_ixs)
    return self.tgt_data[ix]

  def __getitem__(self, ix: int) -> Item_t:
    src = self.src_data[ix]
    src = {f"{self.SRC_PREFIX}_{k}": v for k, v in src.items()}

    tgt = self.tgt_data[ix] if self.is_aligned else self._sample_from_target(ix)
    tgt = {f"{self.TGT_PREFIX}_{k}": v for k, v in tgt.items()}

    return {**src, **tgt}

  def __len__(self) -> int:
    return len(self.src_data)



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any, Callable, Optional, Sequence, Tuple, Union

import jax
import jax.numpy as jnp

from flax import linen as nn

from ott.neural.networks import potentials
from ott.neural.networks.layers import posdef

__all__ = ["ICNN"]

DEFAULT_KERNEL_INIT = posdef.DEFAULT_KERNEL_INIT
DEFAULT_RECTIFIER = nn.activation.relu
DEFAULT_ACTIVATION = nn.activation.relu


class ICNN(potentials.BasePotential):
  """Input convex neural network (ICNN).

  Implementation of input convex neural networks as introduced in
  :cite:`amos:17` with initialization schemes proposed by :cite:`bunne:22`,
  and (low-rank + diagonal) quadratic on inputs at each layer, by
  :cite:`vesseron:24`.

  Args:
    dim_data: data dimensionality.
    dim_hidden: sequence specifying size of hidden dimensions. The
      output dimension of the last layer is 1 by default.
    ranks: ranks of the matrices :math:`A_i` used as low-rank factors
      for the quadratic potentials. If a sequence is passed, it must contain
      ``len(dim_hidden) + 2`` elements, where the last 2 elements correspond
      to the ranks of the final layer with dimension 1 and the potentials,
      respectively.
    init_fn: Initializer for the kernel weight matrices.
      The default is :func:`~flax.linen.initializers.normal`.
    act_fn: choice of activation function used in network architecture,
      needs to be convex. The default is :func:`~flax.linen.activation.relu`.
    pos_weights: Enforce positive weights with a projection.
      If :obj:`False`, the positive weights should be enforced with clipping
      or regularization in the loss.
    rectifier_fn: function to ensure the non negativity of the weights.
      The default is :func:`~flax.linen.activation.relu`.
    gaussian_map_samples: Tuple of source and target points, used to initialize
      the ICNN to mimic the linear Bures map that morphs the (Gaussian
      approximation) of the input measure to that of the target measure. If
      :obj:`None`, the identity initialization is used, and ICNN mimics half the
      squared Euclidean norm.
  """

  dim_data: int
  dim_hidden: Sequence[int]
  ranks: Union[int, Tuple[int, ...]] = 1
  init_fn: Callable[[jax.Array, Tuple[int, ...], Any],
                    jnp.ndarray] = DEFAULT_KERNEL_INIT
  act_fn: Callable[[jnp.ndarray], jnp.ndarray] = DEFAULT_ACTIVATION
  pos_weights: bool = False
  rectifier_fn: Callable[[jnp.ndarray], jnp.ndarray] = DEFAULT_RECTIFIER
  gaussian_map_samples: Optional[Tuple[jnp.ndarray, jnp.ndarray]] = None

  def setup(self) -> None:  # noqa: D102
    dim_hidden = list(self.dim_hidden) + [1]
    *ranks, pos_def_rank = self._normalize_ranks()

    # final layer computes average, still with normalized rescaling
    self.w_zs = [self._get_wz(dim) for dim in dim_hidden[1:]]
    # subsequent layers re-injected into convex functions
    self.w_xs = [
        self._get_wx(dim, rank) for dim, rank in zip(dim_hidden, ranks)
    ]
    self.pos_def_potentials = self._get_pos_def_potentials(pos_def_rank)

  @nn.compact
  def __call__(self, x: jnp.ndarray) -> float:  # noqa: D102
    w_x, *w_xs = self.w_xs
    assert len(self.w_zs) == len(w_xs), (len(self.w_zs), len(w_xs))

    z = self.act_fn(w_x(x))
    for w_z, w_x in zip(self.w_zs, w_xs):
      z = self.act_fn(w_z(z) + w_x(x))
    z = z + self.pos_def_potentials(x)

    return z.squeeze()

  def _get_wz(self, dim: int) -> nn.Module:
    if self.pos_weights:
      return posdef.PositiveDense(
          dim,
          kernel_init=self.init_fn,
          use_bias=False,
          rectifier_fn=self.rectifier_fn,
      )

    return nn.Dense(
        dim,
        kernel_init=self.init_fn,
        use_bias=False,
    )

  def _get_wx(self, dim: int, rank: int) -> nn.Module:
    return posdef.PosDefPotentials(
        rank=rank,
        num_potentials=dim,
        use_linear=True,
        use_bias=True,
        kernel_diag_init=nn.initializers.constant(-2.0),
        rectifier_fn=jax.nn.softplus,
        kernel_lr_init=self.init_fn,
        kernel_linear_init=self.init_fn,
        bias_init=nn.initializers.zeros,
    )

  def _get_pos_def_potentials(self, rank: int) -> posdef.PosDefPotentials:
    kwargs = {
        "num_potentials": 1,
        "use_linear": True,
        "use_bias": True,
        "bias_init": nn.initializers.zeros
    }

    if self.gaussian_map_samples is None:
      return posdef.PosDefPotentials(
          rank=rank,
          kernel_diag_init=nn.initializers.ones,
          kernel_lr_init=nn.initializers.zeros,
          kernel_linear_init=nn.initializers.zeros,
          **kwargs,
      )

    source, target = self.gaussian_map_samples
    return posdef.PosDefPotentials.init_from_samples(
        source,
        target,
        rank=self.dim_data,
        kernel_diag_init=nn.initializers.zeros,
        **kwargs,
    )

  def _normalize_ranks(self) -> Tuple[int, ...]:
    # +2 for the newly added layer with 1 + the final potentials
    n_ranks = len(self.dim_hidden) + 2
    if isinstance(self.ranks, int):
      return (self.ranks,) * n_ranks

    assert len(self.ranks) == n_ranks, (len(self.ranks), n_ranks)
    return tuple(self.ranks)

  @property
  def is_potential(self) -> bool:  # noqa: D102
    return True



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import abc
from typing import Any, Dict, Optional, Sequence, Tuple

import jax
import jax.numpy as jnp

from ott import utils
from ott.geometry import pointcloud
from ott.problems.linear import linear_problem

__all__ = [
    "DefaultInitializer", "GaussianInitializer", "SortingInitializer",
    "SubsampleInitializer"
]


@jax.tree_util.register_pytree_node_class
class SinkhornInitializer(abc.ABC):
  """Base class for Sinkhorn initializers."""

  @abc.abstractmethod
  def init_dual_a(
      self,
      ot_prob: linear_problem.LinearProblem,
      lse_mode: bool,
      rng: Optional[jax.Array] = None,
  ) -> jnp.ndarray:
    """Initialize Sinkhorn potential/scaling f_u.

    Args:
      ot_prob: Linear OT problem.
      lse_mode: Return potential if ``True``, scaling if ``False``.
      rng: Random number generator for stochastic initializers.

    Returns:
      potential/scaling, array of size ``[n,]``.
    """

  @abc.abstractmethod
  def init_dual_b(
      self,
      ot_prob: linear_problem.LinearProblem,
      lse_mode: bool,
      rng: Optional[jax.Array] = None,
  ) -> jnp.ndarray:
    """Initialize Sinkhorn potential/scaling g_v.

    Args:
      ot_prob: Linear OT problem.
      lse_mode: Return potential if ``True``, scaling if ``False``.
      rng: Random number generator for stochastic initializers.

    Returns:
      potential/scaling, array of size ``[m,]``.
    """

  def __call__(
      self,
      ot_prob: linear_problem.LinearProblem,
      a: Optional[jnp.ndarray],
      b: Optional[jnp.ndarray],
      lse_mode: bool,
      rng: Optional[jax.Array] = None,
  ) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """Initialize Sinkhorn potentials/scalings f_u and g_v.

    Args:
      ot_prob: Linear OT problem.
      a: Initial potential/scaling f_u.
        If ``None``, it will be initialized using :meth:`init_dual_a`.
      b: Initial potential/scaling g_v.
        If ``None``, it will be initialized using :meth:`init_dual_b`.
      lse_mode: Return potentials if ``True``, scalings if ``False``.
      rng: Random number generator for stochastic initializers.

    Returns:
      The initial potentials/scalings.
    """
    rng = utils.default_prng_key(rng)
    rng_x, rng_y = jax.random.split(rng, 2)
    n, m = ot_prob.geom.shape
    if a is None:
      a = self.init_dual_a(ot_prob, lse_mode=lse_mode, rng=rng_x)
    if b is None:
      b = self.init_dual_b(ot_prob, lse_mode=lse_mode, rng=rng_y)

    assert a.shape == (
        n,
    ), f"Expected `f_u` to have shape `{n,}`, found `{a.shape}`."
    assert b.shape == (
        m,
    ), f"Expected `g_v` to have shape `{m,}`, found `{b.shape}`."

    # cancel dual variables for zero weights
    a = jnp.where(ot_prob.a > 0.0, a, -jnp.inf if lse_mode else 0.0)
    b = jnp.where(ot_prob.b > 0.0, b, -jnp.inf if lse_mode else 0.0)

    return a, b

  def tree_flatten(self) -> Tuple[Sequence[Any], Dict[str, Any]]:  # noqa: D102
    return [], {}

  @classmethod
  def tree_unflatten(  # noqa: D102
      cls, aux_data: Dict[str, Any], children: Sequence[Any]
  ) -> "SinkhornInitializer":
    return cls(*children, **aux_data)


@jax.tree_util.register_pytree_node_class
class DefaultInitializer(SinkhornInitializer):
  """Default initialization of Sinkhorn dual potentials/primal scalings."""

  def init_dual_a(  # noqa: D102
      self,
      ot_prob: linear_problem.LinearProblem,
      lse_mode: bool,
      rng: Optional[jax.Array] = None,
  ) -> jnp.ndarray:
    del rng
    return jnp.zeros_like(ot_prob.a) if lse_mode else jnp.ones_like(ot_prob.a)

  def init_dual_b(  # noqa: D102
      self,
      ot_prob: linear_problem.LinearProblem,
      lse_mode: bool,
      rng: Optional[jax.Array] = None,
  ) -> jnp.ndarray:
    del rng
    return jnp.zeros_like(ot_prob.b) if lse_mode else jnp.ones_like(ot_prob.b)


@jax.tree_util.register_pytree_node_class
class GaussianInitializer(DefaultInitializer):
  """Gaussian initializer :cite:`thornton2022rethinking:22`.

  Compute Gaussian approximations of each
  :class:`~ott.geometry.pointcloud.PointCloud`, then compute closed from
  Kantorovich potential between Gaussian approximations using Brenier's theorem
  (adapt convex/Brenier potential to Kantorovich). Use this Gaussian potential
  to initialize Sinkhorn potentials/scalings.
  """

  def init_dual_a(  # noqa: D102
      self,
      ot_prob: linear_problem.LinearProblem,
      lse_mode: bool,
      rng: Optional[jax.Array] = None,
  ) -> jnp.ndarray:
    # import Gaussian here due to circular imports
    from ott.tools.gaussian_mixture import gaussian

    del rng
    assert isinstance(
        ot_prob.geom, pointcloud.PointCloud
    ), "Gaussian initializer valid only for pointcloud geoms."

    x, y = ot_prob.geom.x, ot_prob.geom.y
    a, b = ot_prob.a, ot_prob.b

    gaussian_a = gaussian.Gaussian.from_samples(x, weights=a)
    gaussian_b = gaussian.Gaussian.from_samples(y, weights=b)
    # Brenier potential for cost ||x-y||^2/2, multiply by two for ||x-y||^2
    f_potential = 2 * gaussian_a.f_potential(dest=gaussian_b, points=x)
    f_potential = f_potential - jnp.mean(f_potential)
    return f_potential if lse_mode else ot_prob.geom.scaling_from_potential(
        f_potential
    )


@jax.tree_util.register_pytree_node_class
class SortingInitializer(DefaultInitializer):
  """Sorting initializer :cite:`thornton2022rethinking:22`.

  Solve non-regularized OT problem via sorting, then compute potential through
  iterated minimum on C-transform and use this potential to initialize
  regularized potential.

  Args:
    vectorized_update: Whether to use vectorized loop.
    tolerance: DualSort convergence threshold.
    max_iter: Max DualSort steps.
  """

  def __init__(
      self,
      vectorized_update: bool = True,
      tolerance: float = 1e-2,
      max_iter: int = 100
  ):
    super().__init__()
    self.tolerance = tolerance
    self.max_iter = max_iter
    self.vectorized_update = vectorized_update

  def _init_sorting_dual(
      self, modified_cost: jnp.ndarray, init_f: jnp.ndarray
  ) -> jnp.ndarray:
    """Run DualSort algorithm.

    Args:
      modified_cost: cost matrix minus diagonal column-wise.
      init_f: potential f, array of size n. This is the starting potential,
        which is then updated to make the init potential, so an init of an init.

    Returns:
      potential f, array of size n.
    """

    def body_fn(
        state: Tuple[jnp.ndarray, float, int]
    ) -> Tuple[jnp.ndarray, float, int]:
      prev_f, _, it = state
      new_f = fn(prev_f, modified_cost)
      diff = jnp.sum((new_f - prev_f) ** 2)
      it += 1
      return new_f, diff, it

    def cond_fn(state: Tuple[jnp.ndarray, float, int]) -> bool:
      _, diff, it = state
      return jnp.logical_and(diff > self.tolerance, it < self.max_iter)

    fn = _vectorized_update if self.vectorized_update else _coordinate_update
    state = (init_f, jnp.inf, 0)  # init, error, iter
    f_potential, _, _ = jax.lax.while_loop(
        cond_fun=cond_fn, body_fun=body_fn, init_val=state
    )

    return f_potential

  def init_dual_a(
      self,
      ot_prob: linear_problem.LinearProblem,
      lse_mode: bool,
      rng: Optional[jax.Array] = None,
      init_f: Optional[jnp.ndarray] = None,
  ) -> jnp.ndarray:
    """Apply DualSort algorithm.

    Args:
      ot_prob: OT problem between discrete distributions.
      lse_mode: Return potential if ``True``, scaling if ``False``.
      rng: Random number generator for stochastic initializers, unused.
      init_f: potential f, array of size ``[n,]``. This is the starting
        potential, which is then updated to make the init potential,
        so an init of an init.

    Returns:
      potential/scaling f_u, array of size ``[n,]``.
    """
    del rng
    assert not ot_prob.geom.is_online, \
        "Sorting initializer does not work for online geometry."
    # check for sorted x, y requires point cloud and could slow initializer
    cost_matrix = ot_prob.geom.cost_matrix

    assert cost_matrix.shape[0] == cost_matrix.shape[
        1], "Requires square cost matrix."

    modified_cost = cost_matrix - jnp.diag(cost_matrix)[None, :]

    n = cost_matrix.shape[0]
    init_f = jnp.zeros(n, dtype=ot_prob.dtype) if init_f is None else init_f

    f_potential = self._init_sorting_dual(modified_cost, init_f)
    f_potential = f_potential - jnp.mean(f_potential)

    return f_potential if lse_mode else ot_prob.geom.scaling_from_potential(
        f_potential
    )

  def tree_flatten(self) -> Tuple[Sequence[Any], Dict[str, Any]]:  # noqa: D102
    return ([], {
        "tolerance": self.tolerance,
        "max_iter": self.max_iter,
        "vectorized_update": self.vectorized_update
    })


@jax.tree_util.register_pytree_node_class
class SubsampleInitializer(DefaultInitializer):
  """Subsample initializer :cite:`thornton2022rethinking:22`.

  Subsample each :class:`~ott.geometry.pointcloud.PointCloud`, then compute
  :class:`Sinkhorn potential <ott.problems.linear.potentials.DualPotentials>`
  from the subsampled approximations and use this potential to initialize
  Sinkhorn potentials/scalings for the original problem.

  Args:
    subsample_n_x: number of points to subsample from the first measure in
      :class:`~ott.geometry.pointcloud.PointCloud`.
    subsample_n_y: number of points to subsample from the second measure in
      :class:`~ott.geometry.pointcloud.PointCloud`.
      If ``None``, use ``subsample_n_x``.
    kwargs: Keyword arguments for
      :class:`~ott.solvers.linear.sinkhorn.Sinkhorn`.
  """

  def __init__(
      self,
      subsample_n_x: int,
      subsample_n_y: Optional[int] = None,
      **kwargs: Any,
  ):
    super().__init__()
    self.subsample_n_x = subsample_n_x
    self.subsample_n_y = subsample_n_y or subsample_n_x
    self.sinkhorn_kwargs = kwargs

  def init_dual_a(  # noqa: D102
      self,
      ot_prob: linear_problem.LinearProblem,
      lse_mode: bool,
      rng: Optional[jax.Array] = None,
  ) -> jnp.ndarray:
    from ott.solvers import linear

    assert isinstance(
        ot_prob.geom, pointcloud.PointCloud
    ), "Subsample initializer valid only for pointcloud geom."
    rng = utils.default_prng_key(rng)
    rng_x, rng_y = jax.random.split(rng, 2)

    x, y = ot_prob.geom.x, ot_prob.geom.y
    a, b = ot_prob.a, ot_prob.b

    # subsample
    sub_x = jax.random.choice(
        rng_x, a=x, shape=(self.subsample_n_x,), replace=True, p=a, axis=0
    )
    sub_y = jax.random.choice(
        rng_y, a=y, shape=(self.subsample_n_y,), replace=True, p=b, axis=0
    )

    # create subsampled point cloud geometry
    sub_geom = pointcloud.PointCloud(
        sub_x,
        sub_y,
        epsilon=ot_prob.geom.epsilon,
        scale_cost=ot_prob.geom._scale_cost,
        cost_fn=ot_prob.geom.cost_fn
    )

    # run sinkhorn
    subsample_sink_out = linear.solve(sub_geom, **self.sinkhorn_kwargs)

    # interpolate potentials
    dual_potentials = subsample_sink_out.to_dual_potentials()
    f_potential = jax.vmap(dual_potentials.f)(x)

    return f_potential if lse_mode else ot_prob.geom.scaling_from_potential(
        f_potential
    )

  def tree_flatten(self) -> Tuple[Sequence[Any], Dict[str, Any]]:  # noqa: D102
    return ([], {
        "subsample_n_x": self.subsample_n_x,
        "subsample_n_y": self.subsample_n_y,
        **self.sinkhorn_kwargs
    })


def _vectorized_update(
    f: jnp.ndarray, modified_cost: jnp.ndarray
) -> jnp.ndarray:
  """Inner loop DualSort Update.

  Args:
    f: potential f, array of size n.
    modified_cost: cost matrix minus diagonal column-wise.

  Returns:
    updated potential vector, f.
  """
  return jnp.min(modified_cost + f[None, :], axis=1)


def _coordinate_update(
    f: jnp.ndarray, modified_cost: jnp.ndarray
) -> jnp.ndarray:
  """Coordinate-wise updates within inner loop.

  Args:
    f: potential f, array of size n.
    modified_cost: cost matrix minus diagonal column-wise.

  Returns:
    updated potential vector, f.
  """

  def body_fn(i: int, f: jnp.ndarray) -> jnp.ndarray:
    new_f = jnp.min(modified_cost[i, :] + f)
    return f.at[i].set(new_f)

  return jax.lax.fori_loop(0, len(f), body_fn, f)



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any, Dict, Optional, Sequence, Tuple

import jax
import jax.numpy as jnp

from ott.geometry import costs, geometry, segment

__all__ = ["FreeBarycenterProblem", "FixedBarycenterProblem"]


@jax.tree_util.register_pytree_node_class
class FreeBarycenterProblem:
  """Free Wasserstein barycenter problem :cite:`cuturi:14`.

  Args:
    y: Array of shape ``[num_total_points, ndim]`` merging the points of all
      measures. Alternatively, already segmented array of shape
      ``[num_measures, max_measure_size, ndim]`` can be passed.
      See also :func:`~ott.geometry.segment.segment_point_cloud`.
    b: Array of shape ``[num_total_points,]`` containing the weights of all
      the points within the measures that define the barycenter problem.
      Same as ``y``, pre-segmented array of weights of shape
      ``[num_measures, max_measure_size]`` can be passed.
      If ``y`` is already pre-segmented, this array must be always specified.
    weights: Array of shape ``[num_measures,]`` containing the weights of the
      measures.
    cost_fn: Cost function used. If `None`,
      use the :class:`~ott.geometry.costs.SqEuclidean` cost.
    epsilon: Epsilon regularization used to solve reg-OT problems.
    kwargs: Keyword arguments :func:`~ott.geometry.segment.segment_point_cloud`.
      Only used when ``y`` is not already segmented. When passing
      ``segment_ids``, 2 arguments must be specified for jitting to work:

      - ``num_segments`` - the total number of measures.
      - ``max_measure_size`` -  maximum of support sizes of these measures.
  """

  def __init__(
      self,
      y: jnp.ndarray,
      b: Optional[jnp.ndarray] = None,
      weights: Optional[jnp.ndarray] = None,
      cost_fn: Optional[costs.CostFn] = None,
      epsilon: Optional[float] = None,
      **kwargs: Any,
  ):
    self._y = y
    if y.ndim == 3 and b is None:
      raise ValueError("Specify weights if `y` is already segmented.")
    self._b = b
    self._weights = weights
    self.cost_fn = costs.SqEuclidean() if cost_fn is None else cost_fn
    self.epsilon = epsilon
    self._kwargs = kwargs

    if self._is_segmented:
      # (num_measures, max_measure_size, ndim)
      # (num_measures, max_measure_size)
      assert self._y.shape[:2] == self._b.shape
    else:
      # (num_total_points, ndim)
      # (num_total_points,)
      assert self._b is None or self._y.shape[0] == self._b.shape[0]

  @property
  def segmented_y_b(self) -> Tuple[jnp.ndarray, jnp.ndarray]:
    """Tuple of arrays containing the segmented measures and weights.

    - Segmented measures of shape ``[num_measures, max_measure_size, ndim]``.
    - Segmented weights of shape ``[num_measures, max_measure_size]``.
    """
    if self._is_segmented:
      y, b = self._y, self._b
    else:
      y, b = segment.segment_point_cloud(
          x=self._y,
          a=self._b,
          padding_vector=self.cost_fn._padder(self.ndim),
          **self._kwargs
      )
    return y, b

  @property
  def flattened_y(self) -> jnp.ndarray:
    """Array of shape ``[num_measures * (N_1 + N_2 + ...), ndim]``."""
    if self._is_segmented:
      return self._y.reshape((-1, self._y.shape[-1]))
    return self._y

  @property
  def flattened_b(self) -> Optional[jnp.ndarray]:
    """Array of shape ``[num_measures * (N_1 + N_2 + ...),]``."""
    return None if self._b is None else self._b.ravel()

  @property
  def num_measures(self) -> int:
    """Number of measures."""
    return self.segmented_y_b[0].shape[0]

  @property
  def max_measure_size(self) -> int:
    """Maximum number of points across all measures."""
    return self.segmented_y_b[0].shape[1]

  @property
  def ndim(self) -> int:
    """Number of dimensions of ``y``."""
    return self._y.shape[-1]

  @property
  def weights(self) -> jnp.ndarray:
    """Barycenter weights of shape ``[num_measures,]`` that sum to 1."""
    if self._weights is None:
      return jnp.ones((self.num_measures,)) / self.num_measures
    # Check that the number of measures coincides with the weights' size.
    assert self._weights.shape[0] == self.num_measures
    # By default, we assume that weights sum to 1, and enforce this if needed.
    return self._weights / jnp.sum(self._weights)

  @property
  def _is_segmented(self) -> bool:
    return self._y.ndim == 3

  def tree_flatten(self) -> Tuple[Sequence[Any], Dict[str, Any]]:  # noqa: D102
    return ([self._y, self._b, self._weights], {
        "cost_fn": self.cost_fn,
        "epsilon": self.epsilon,
        **self._kwargs,
    })

  @classmethod
  def tree_unflatten(  # noqa: D102
      cls, aux_data: Dict[str, Any], children: Sequence[Any]
  ) -> "FreeBarycenterProblem":
    y, b, weights = children
    return cls(y=y, b=b, weights=weights, **aux_data)


@jax.tree_util.register_pytree_node_class
class FixedBarycenterProblem:
  """Fixed-support Wasserstein barycenter problem.

  Args:
    geom: Geometry object.
    a: batch of histograms of shape ``[batch, num_a]`` where ``num_a`` matches
      the first value of the :attr:`~ott.geometry.Geometry.shape` attribute of
      ``geom``.
    weights: ``[batch,]`` positive weights summing to :math:`1`. Uniform by
      default.
  """

  def __init__(
      self,
      geom: geometry.Geometry,
      a: jnp.ndarray,
      weights: Optional[jnp.ndarray] = None,
  ):
    self.geom = geom
    self.a = a
    self._weights = weights

  @property
  def num_measures(self) -> int:
    """Number of measures."""
    return self.a.shape[0]

  @property
  def weights(self) -> jnp.ndarray:
    """Barycenter weights of shape ``[num_measures,]`` that sum to :math`1`."""
    if self._weights is None:
      return jnp.ones((self.num_measures,)) / self.num_measures

    # check that the number of measures coincides with the weights' size
    assert self._weights.shape[0] == self.num_measures
    # by default, we assume that weights sum to 1, and enforce this if needed
    return self._weights / jnp.sum(self._weights)

  def tree_flatten(self):  # noqa: D102
    return [self.geom, self.a, self._weights], None

  @classmethod
  def tree_unflatten(  # noqa: D102
      cls, aux_data: Dict[str, Any], children: Sequence[Any]
  ) -> "FixedBarycenterProblem":
    del aux_data
    geom, a, weights = children
    return cls(geom=geom, a=a, weights=weights)



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import dataclasses
from typing import Iterator, Literal, NamedTuple, Optional, Tuple

import jax
import jax.numpy as jnp
import numpy as np

__all__ = ["create_gaussian_mixture_samplers", "Dataset", "GaussianMixture"]

from ott import utils

Name_t = Literal["simple", "circle", "square_five", "square_four"]


class Dataset(NamedTuple):
  r"""Samplers from source and target measures.

  Args:
    source_iter: loader for the source measure
    target_iter: loader for the target measure
  """
  source_iter: Iterator[jnp.ndarray]
  target_iter: Iterator[jnp.ndarray]


@dataclasses.dataclass
class GaussianMixture:
  """A mixture of Gaussians.

  Args:
    name: the name specifying the centers of the mixture components:

      - ``simple`` - data clustered in one center,
      - ``circle`` - two-dimensional Gaussians arranged on a circle,
      - ``square_five`` - two-dimensional Gaussians on a square with
        one Gaussian in the center, and
      - ``square_four`` - two-dimensional Gaussians in the corners of a
        rectangle

    batch_size: batch size of the samples
    rng: initial PRNG key
    scale: scale of the Gaussian means
    std: the standard deviation of the individual Gaussian samples
  """
  name: Name_t
  batch_size: int
  rng: jax.Array
  scale: float = 5.0
  std: float = 0.5

  def __post_init__(self) -> None:
    gaussian_centers = {
        "simple":
            np.array([[0, 0]]),
        "circle":
            np.array([
                (1, 0),
                (-1, 0),
                (0, 1),
                (0, -1),
                (1.0 / np.sqrt(2), 1.0 / np.sqrt(2)),
                (1.0 / np.sqrt(2), -1.0 / np.sqrt(2)),
                (-1.0 / np.sqrt(2), 1.0 / np.sqrt(2)),
                (-1.0 / np.sqrt(2), -1.0 / np.sqrt(2)),
            ]),
        "square_five":
            np.array([[0, 0], [1, 1], [-1, 1], [-1, -1], [1, -1]]),
        "square_four":
            np.array([[1, 0], [0, 1], [-1, 0], [0, -1]]),
    }
    if self.name not in gaussian_centers:
      raise ValueError(
          f"{self.name} is not a valid dataset for GaussianMixture"
      )
    self.centers = gaussian_centers[self.name]

  def __iter__(self) -> Iterator[jnp.array]:
    """Random sample generator from Gaussian mixture.

    Returns:
      A generator of samples from the Gaussian mixture.
    """
    return self._create_sample_generators()

  def _create_sample_generators(self) -> Iterator[jnp.array]:
    rng = self.rng
    while True:
      rng1, rng2, rng = jax.random.split(rng, 3)
      means = jax.random.choice(rng1, self.centers, (self.batch_size,))
      normal_samples = jax.random.normal(rng2, (self.batch_size, 2))
      samples = self.scale * means + (self.std ** 2) * normal_samples
      yield samples


def create_gaussian_mixture_samplers(
    name_source: Name_t,
    name_target: Name_t,
    train_batch_size: int = 2048,
    valid_batch_size: int = 2048,
    rng: Optional[jax.Array] = None,
) -> Tuple[Dataset, Dataset, int]:
  """Gaussian samplers.

  Args:
    name_source: name of the source sampler
    name_target: name of the target sampler
    train_batch_size: the training batch size
    valid_batch_size: the validation batch size
    rng: initial PRNG key

  Returns:
    The dataset and dimension of the data.
  """
  rng = utils.default_prng_key(rng)
  rng1, rng2, rng3, rng4 = jax.random.split(rng, 4)
  train_dataset = Dataset(
      source_iter=iter(
          GaussianMixture(name_source, batch_size=train_batch_size, rng=rng1)
      ),
      target_iter=iter(
          GaussianMixture(name_target, batch_size=train_batch_size, rng=rng2)
      )
  )
  valid_dataset = Dataset(
      source_iter=iter(
          GaussianMixture(name_source, batch_size=valid_batch_size, rng=rng3)
      ),
      target_iter=iter(
          GaussianMixture(name_target, batch_size=valid_batch_size, rng=rng4)
      )
  )
  dim_data = 2
  return train_dataset, valid_dataset, dim_data



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import TYPE_CHECKING, Any, Dict, Sequence, Tuple, Union

import jax
import jax.numpy as jnp

from ott.solvers.linear import sinkhorn, sinkhorn_lr

if TYPE_CHECKING:
  from ott.solvers.linear import continuous_barycenter

__all__ = ["WassersteinSolver"]

State = Union[sinkhorn.SinkhornState, sinkhorn_lr.LRSinkhornState,
              "continuous_barycenter.FreeBarycenterState"]


@jax.tree_util.register_pytree_node_class
class WassersteinSolver:
  """A generic solver for problems that use a linear problem in inner loop."""

  def __init__(
      self,
      linear_solver: Union["sinkhorn.Sinkhorn", "sinkhorn_lr.LRSinkhorn"],
      threshold: float = 1e-3,
      min_iterations: int = 5,
      max_iterations: int = 50,
      store_inner_errors: bool = False,
  ):
    self.linear_solver = linear_solver
    self.min_iterations = min_iterations
    self.max_iterations = max_iterations
    self.threshold = threshold
    self.store_inner_errors = store_inner_errors

  @property
  def rank(self) -> int:
    """Rank of the linear OT solver."""
    return self.linear_solver.rank if self.is_low_rank else -1

  @property
  def is_low_rank(self) -> bool:
    """Whether the solver is low-rank."""
    return isinstance(self.linear_solver, sinkhorn_lr.LRSinkhorn)

  def tree_flatten(self) -> Tuple[Sequence[Any], Dict[str, Any]]:  # noqa: D102
    return ([self.linear_solver, self.threshold], {
        "min_iterations": self.min_iterations,
        "max_iterations": self.max_iterations,
        "store_inner_errors": self.store_inner_errors,
    })

  @classmethod
  def tree_unflatten(  # noqa: D102
      cls, aux_data: Dict[str, Any], children: Sequence[Any]
  ) -> "WassersteinSolver":
    return cls(*children, **aux_data)

  def _converged(self, state: State, iteration: int) -> bool:
    costs, i, tol = state.costs, iteration, self.threshold
    return jnp.logical_and(
        i >= 2, jnp.isclose(costs[i - 2], costs[i - 1], rtol=tol)
    )

  def _diverged(self, state: State, iteration: int) -> bool:
    return jnp.logical_not(jnp.isfinite(state.costs[iteration - 1]))

  def _continue(self, state: State, iteration: int) -> bool:
    """Continue while not(converged) and not(diverged)."""
    return jnp.logical_or(
        iteration <= 2,
        jnp.logical_and(
            jnp.logical_not(self._diverged(state, iteration)),
            jnp.logical_not(self._converged(state, iteration))
        )
    )



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import abc

import jax
import jax.numpy as jnp

__all__ = [
    "BaseFlow",
    "StraightFlow",
    "ConstantNoiseFlow",
    "BrownianBridge",
]


class BaseFlow(abc.ABC):
  """Base class for all flows.

  Args:
    sigma: Noise used for computing time-dependent noise schedule.
  """

  def __init__(self, sigma: float):
    self.sigma = sigma

  @abc.abstractmethod
  def compute_mu_t(
      self, t: jnp.ndarray, src: jnp.ndarray, tgt: jnp.ndarray
  ) -> jnp.ndarray:
    """Compute the mean of the probability path.

    Compute the mean of the probability path between :math:`x_0` and :math:`x_1`
    at time :math:`t`.

    Args:
      t: Time :math:`t` of shape ``[batch, 1]``.
      src: Sample from the source distribution of shape ``[batch, ...]``.
      tgt: Sample from the target distribution of shape ``[batch, ...]``.
    """

  @abc.abstractmethod
  def compute_sigma_t(self, t: jnp.ndarray) -> jnp.ndarray:
    """Compute the standard deviation of the probability path at time :math:`t`.

    Args:
      t: Time :math:`t` of shape ``[batch, 1]``.

    Returns:
      Standard deviation of the probability path at time :math:`t`.
    """

  @abc.abstractmethod
  def compute_ut(
      self, t: jnp.ndarray, src: jnp.ndarray, tgt: jnp.ndarray
  ) -> jnp.ndarray:
    """Evaluate the conditional vector field.

    Evaluate the conditional vector field defined between :math:`x_0` and
    :math:`x_1` at time :math:`t`.

    Args:
      t: Time :math:`t` of shape ``[batch, 1]``.
      src: Sample from the source distribution of shape ``[batch, ...]``.
      tgt: Sample from the target distribution of shape ``[batch, ...]``.

    Returns:
      Conditional vector field evaluated at time :math:`t`.
    """

  def compute_xt(
      self, rng: jax.Array, t: jnp.ndarray, src: jnp.ndarray, tgt: jnp.ndarray
  ) -> jnp.ndarray:
    """Sample from the probability path.

    Sample from the probability path between :math:`x_0` and :math:`x_1` at
    time :math:`t`.

    Args:
      rng: Random number generator.
      t: Time :math:`t` of shape ``[batch, 1]``.
      src: Sample from the source distribution of shape ``[batch, ...]``.
      tgt: Sample from the target distribution of shape ``[batch, ...]``.

    Returns:
      Samples from the probability path between :math:`x_0` and :math:`x_1`
      at time :math:`t`.
    """
    noise = jax.random.normal(rng, shape=src.shape)
    mu_t = self.compute_mu_t(t, src, tgt)
    sigma_t = self.compute_sigma_t(t)
    return mu_t + sigma_t * noise


class StraightFlow(BaseFlow, abc.ABC):
  """Base class for flows with straight paths.

  Args:
    sigma: Noise used for computing time-dependent noise schedule.
  """

  def compute_mu_t(  # noqa: D102
      self, t: jnp.ndarray, src: jnp.ndarray, tgt: jnp.ndarray
  ) -> jnp.ndarray:
    return (1.0 - t) * src + t * tgt

  def compute_ut(  # noqa: D102
      self, t: jnp.ndarray, src: jnp.ndarray, tgt: jnp.ndarray
  ) -> jnp.ndarray:
    del t
    return tgt - src


class ConstantNoiseFlow(StraightFlow):
  r"""Flow with straight paths and constant flow noise :math:`\sigma`.

  Args:
    sigma: Constant noise used for computing time-independent noise schedule.
  """

  def compute_sigma_t(self, t: jnp.ndarray) -> jnp.ndarray:
    r"""Compute noise of the flow at time :math:`t`.

    Args:
      t: Time :math:`t` of shape ``[batch, 1]``.

    Returns:
      Constant, time-independent standard deviation :math:`\sigma`.
    """
    return jnp.full_like(t, fill_value=self.sigma)


class BrownianBridge(StraightFlow):
  r"""Brownian Bridge.

  Sampler for sampling noise implicitly defined by a Schrödinger Bridge
  problem with parameter :math:`\sigma` such that
  :math:`\sigma_t = \sigma \cdot \sqrt{t \cdot (1 - t)}` :cite:`tong:23`.

  Args:
    sigma: Noise used for computing time-dependent noise schedule.
  """

  def compute_sigma_t(self, t: jnp.ndarray) -> jnp.ndarray:
    r"""Compute noise of the flow at time :math:`t`.

    Args:
      t: Time :math:`t` of shape ``[batch, 1]``.

    Returns:
      Samples from the probability path between :math:`x_0` and :math:`x_1`
      at time :math:`t`.
    """
    return self.sigma * jnp.sqrt(t * (1.0 - t))



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import functools
from typing import Any, Callable, Dict, Iterable, List, Optional, Tuple, Union

import jax
import jax.numpy as jnp
import jax.tree_util as jtu
import numpy as np

import diffrax
from flax.training import train_state

from ott import utils
from ott.neural.methods.flows import dynamics
from ott.neural.networks import velocity_field
from ott.solvers import utils as solver_utils

__all__ = ["GENOT"]

LinTerm = Tuple[jnp.ndarray, jnp.ndarray]
QuadTerm = Tuple[jnp.ndarray, jnp.ndarray, Optional[jnp.ndarray],
                 Optional[jnp.ndarray]]
DataMatchFn = Union[Callable[[LinTerm], jnp.ndarray], Callable[[QuadTerm],
                                                               jnp.ndarray]]


class GENOT:
  """Generative Entropic Neural Optimal Transport :cite:`klein_uscidda:23`.

  GENOT is a framework for learning neural optimal transport plans between
  two distributions. It allows for learning linear and quadratic
  (Fused) Gromov-Wasserstein couplings, in both the balanced and
  the unbalanced setting.

  Args:
    vf: Vector field parameterized by a neural network.
    flow: Flow between the latent and the target distributions.
    data_match_fn: Function to match samples from the source and the target
      distributions. Depending on the data passed in :meth:`__call__`, it has
      the following signature:

      - ``(src_lin, tgt_lin) -> matching`` - linear matching.
      - ``(src_quad, tgt_quad, src_lin, tgt_lin) -> matching`` -
        quadratic (fused) GW matching. In the pure GW setting, both ``src_lin``
        and ``tgt_lin`` will be set to :obj:`None`.

    source_dim: Dimensionality of the source distribution.
    target_dim: Dimensionality of the target distribution.
    condition_dim: Dimension of the conditions. If :obj:`None`, the underlying
      velocity field has no conditions.
    time_sampler: Time sampler with a ``(rng, n_samples) -> time`` signature.
    latent_noise_fn: Function to sample from the latent distribution in the
      target space with a ``(rng, shape) -> noise`` signature.
      If :obj:`None`, multivariate normal distribution is used.
    latent_match_fn: Function to match samples from the latent distribution
      and the samples from the conditional distribution with a
      ``(latent, samples) -> matching`` signature. If :obj:`None`, no matching
      is performed.
    n_samples_per_src: Number of samples drawn from the conditional distribution
      per one source sample.
    kwargs: Keyword arguments for
      :meth:`~ott.neural.networks.velocity_field.VelocityField.create_train_state`.
  """  # noqa: E501

  def __init__(
      self,
      vf: velocity_field.VelocityField,
      flow: dynamics.BaseFlow,
      data_match_fn: DataMatchFn,
      *,
      source_dim: int,
      target_dim: int,
      condition_dim: Optional[int] = None,
      time_sampler: Callable[[jax.Array, int],
                             jnp.ndarray] = solver_utils.uniform_sampler,
      latent_noise_fn: Optional[Callable[[jax.Array, Tuple[int, ...]],
                                         jnp.ndarray]] = None,
      latent_match_fn: Optional[Callable[[jnp.ndarray, jnp.ndarray],
                                         jnp.ndarray]] = None,
      n_samples_per_src: int = 1,
      **kwargs: Any,
  ):
    self.vf = vf
    self.flow = flow
    self.data_match_fn = data_match_fn
    self.time_sampler = time_sampler
    if latent_noise_fn is None:
      latent_noise_fn = functools.partial(_multivariate_normal, dim=target_dim)
    self.latent_noise_fn = latent_noise_fn
    self.latent_match_fn = latent_match_fn
    self.n_samples_per_src = n_samples_per_src

    self.vf_state = self.vf.create_train_state(
        input_dim=target_dim,
        condition_dim=source_dim + (condition_dim or 0),
        **kwargs
    )
    self.step_fn = self._get_step_fn()

  def _get_step_fn(self) -> Callable:

    @jax.jit
    def step_fn(
        rng: jax.Array,
        vf_state: train_state.TrainState,
        time: jnp.ndarray,
        source: jnp.ndarray,
        target: jnp.ndarray,
        latent: jnp.ndarray,
        source_conditions: Optional[jnp.ndarray],
    ):

      def loss_fn(
          params: jnp.ndarray, time: jnp.ndarray, source: jnp.ndarray,
          target: jnp.ndarray, latent: jnp.ndarray,
          source_conditions: Optional[jnp.ndarray], rng: jax.Array
      ) -> jnp.ndarray:
        rng_flow, rng_dropout = jax.random.split(rng, 2)
        x_t = self.flow.compute_xt(rng_flow, time, latent, target)
        if source_conditions is None:
          cond = source
        else:
          cond = jnp.concatenate([source, source_conditions], axis=-1)

        v_t = vf_state.apply_fn({"params": params},
                                time,
                                x_t,
                                cond,
                                rngs={"dropout": rng_dropout})
        u_t = self.flow.compute_ut(time, latent, target)

        return jnp.mean((v_t - u_t) ** 2)

      grad_fn = jax.value_and_grad(loss_fn)
      loss, grads = grad_fn(
          vf_state.params, time, source, target, latent, source_conditions, rng
      )

      return loss, vf_state.apply_gradients(grads=grads)

    return step_fn

  def __call__(
      self,
      loader: Iterable[Dict[str, np.ndarray]],
      n_iters: int,
      rng: Optional[jax.Array] = None
  ) -> Dict[str, List[float]]:
    """Train the GENOT model.

    Args:
      loader: Data loader returning a dictionary with possible keys
        `src_lin`, `tgt_lin`, `src_quad`, `tgt_quad`, `src_conditions`.
      n_iters: Number of iterations to train the model.
      rng: Random key for seeding.

    Returns:
      Training logs.
    """

    def prepare_data(
        batch: Dict[str, jnp.ndarray]
    ) -> Tuple[Tuple[jnp.ndarray, Optional[jnp.ndarray], jnp.ndarray],
               Tuple[jnp.ndarray, jnp.ndarray, Optional[jnp.ndarray],
                     Optional[jnp.ndarray]]]:
      src_lin, src_quad = batch.get("src_lin"), batch.get("src_quad")
      tgt_lin, tgt_quad = batch.get("tgt_lin"), batch.get("tgt_quad")

      if src_quad is None and tgt_quad is None:  # lin
        src, tgt = src_lin, tgt_lin
        arrs = src_lin, tgt_lin
      elif src_lin is None and tgt_lin is None:  # quad
        src, tgt = src_quad, tgt_quad
        arrs = src_quad, tgt_quad
      elif all(
          arr is not None for arr in (src_lin, tgt_lin, src_quad, tgt_quad)
      ):  # fused quad
        src = jnp.concatenate([src_lin, src_quad], axis=1)
        tgt = jnp.concatenate([tgt_lin, tgt_quad], axis=1)
        arrs = src_quad, tgt_quad, src_lin, tgt_lin
      else:
        raise RuntimeError("Cannot infer OT problem type from data.")

      return (src, batch.get("src_condition"), tgt), arrs

    rng = utils.default_prng_key(rng)
    training_logs = {"loss": []}
    for batch in loader:
      rng = jax.random.split(rng, 5)
      rng, rng_resample, rng_noise, rng_time, rng_step_fn = rng

      batch = jtu.tree_map(jnp.asarray, batch)
      (src, src_cond, tgt), matching_data = prepare_data(batch)

      n = src.shape[0]
      time = self.time_sampler(rng_time, n * self.n_samples_per_src)
      latent = self.latent_noise_fn(rng_noise, (n, self.n_samples_per_src))

      tmat = self.data_match_fn(*matching_data)  # (n, m)
      src_ixs, tgt_ixs = solver_utils.sample_conditional(  # (n, k), (m, k)
          rng_resample,
          tmat,
          k=self.n_samples_per_src,
      )

      src, tgt = src[src_ixs], tgt[tgt_ixs]  # (n, k, ...),  # (m, k, ...)
      if src_cond is not None:
        src_cond = src_cond[src_ixs]

      if self.latent_match_fn is not None:
        src, src_cond, tgt = self._match_latent(rng, src, src_cond, latent, tgt)

      src = src.reshape(-1, *src.shape[2:])  # (n * k, ...)
      tgt = tgt.reshape(-1, *tgt.shape[2:])  # (m * k, ...)
      latent = latent.reshape(-1, *latent.shape[2:])
      if src_cond is not None:
        src_cond = src_cond.reshape(-1, *src_cond.shape[2:])

      loss, self.vf_state = self.step_fn(
          rng_step_fn, self.vf_state, time, src, tgt, latent, src_cond
      )

      training_logs["loss"].append(float(loss))
      if len(training_logs["loss"]) >= n_iters:
        break

    return training_logs

  def _match_latent(
      self, rng: jax.Array, src: jnp.ndarray, src_cond: Optional[jnp.ndarray],
      latent: jnp.ndarray, tgt: jnp.ndarray
  ) -> Tuple[jnp.ndarray, Optional[jnp.ndarray], jnp.ndarray]:

    def resample(
        rng: jax.Array, src: jnp.ndarray, src_cond: Optional[jnp.ndarray],
        tgt: jnp.ndarray, latent: jnp.ndarray
    ) -> Tuple[jnp.ndarray, Optional[jnp.ndarray], jnp.ndarray]:
      tmat = self.latent_match_fn(latent, tgt)  # (n, k)

      src_ixs, tgt_ixs = solver_utils.sample_joint(rng, tmat)  # (n,), (m,)
      src, tgt = src[src_ixs], tgt[tgt_ixs]
      if src_cond is not None:
        src_cond = src_cond[src_ixs]

      return src, src_cond, tgt

    cond_axis = None if src_cond is None else 1
    in_axes, out_axes = (0, 1, cond_axis, 1, 1), (1, cond_axis, 1)
    resample_fn = jax.jit(jax.vmap(resample, in_axes, out_axes))

    rngs = jax.random.split(rng, self.n_samples_per_src)
    return resample_fn(rngs, src, src_cond, tgt, latent)

  def transport(
      self,
      source: jnp.ndarray,
      condition: Optional[jnp.ndarray] = None,
      t0: float = 0.0,
      t1: float = 1.0,
      rng: Optional[jax.Array] = None,
      **kwargs: Any,
  ) -> jnp.ndarray:
    """Transport data with the learned plan.

    This function pushes forward the source distribution to its conditional
    distribution by solving the neural ODE.

    Args:
      source: Data to transport.
      condition: Condition of the input data.
      t0: Starting time of integration of neural ODE.
      t1: End time of integration of neural ODE.
      rng: Random generate used to sample from the latent distribution.
      kwargs: Keyword arguments for :func:`~diffrax.odesolve`.

    Returns:
      The push-forward defined by the learned transport plan.
    """

    def vf(t: jnp.ndarray, x: jnp.ndarray, cond: jnp.ndarray) -> jnp.ndarray:
      params = self.vf_state.params
      return self.vf_state.apply_fn({"params": params}, t, x, cond, train=False)

    def solve_ode(x: jnp.ndarray, cond: jnp.ndarray) -> jnp.ndarray:
      ode_term = diffrax.ODETerm(vf)
      sol = diffrax.diffeqsolve(
          ode_term,
          t0=t0,
          t1=t1,
          y0=x,
          args=cond,
          **kwargs,
      )
      return sol.ys[0]

    kwargs.setdefault("dt0", None)
    kwargs.setdefault("solver", diffrax.Tsit5())
    kwargs.setdefault(
        "stepsize_controller", diffrax.PIDController(rtol=1e-5, atol=1e-5)
    )

    rng = utils.default_prng_key(rng)
    latent = self.latent_noise_fn(rng, (len(source),))

    if condition is not None:
      source = jnp.concatenate([source, condition], axis=-1)

    return jax.jit(jax.vmap(solve_ode))(latent, source)


def _multivariate_normal(
    rng: jax.Array,
    shape: Tuple[int, ...],
    dim: int,
    mean: float = 0.0,
    cov: float = 1.0
) -> jnp.ndarray:
  mean = jnp.full(dim, fill_value=mean)
  cov = jnp.diag(jnp.full(dim, fill_value=cov))
  return jax.random.multivariate_normal(rng, mean=mean, cov=cov, shape=shape)



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import collections
import functools
from typing import (
    Any,
    Callable,
    Dict,
    Iterator,
    Literal,
    Optional,
    Sequence,
    Tuple,
    Union,
)

import jax
import jax.numpy as jnp

import optax
from flax.core import frozen_dict
from flax.training import train_state

from ott import utils
from ott.geometry import costs, pointcloud
from ott.neural.networks import potentials
from ott.solvers import linear
from ott.solvers.linear import sinkhorn

__all__ = ["monge_gap", "monge_gap_from_samples", "MongeGapEstimator"]


def monge_gap(
    map_fn: Callable[[jnp.ndarray], jnp.ndarray],
    reference_points: jnp.ndarray,
    cost_fn: Optional[costs.CostFn] = None,
    epsilon: Optional[float] = None,
    relative_epsilon: Optional[bool] = None,
    scale_cost: Union[float, Literal["mean", "max_cost", "median"]] = 1.0,
    return_output: bool = False,
    **kwargs: Any
) -> Union[float, Tuple[float, sinkhorn.SinkhornOutput]]:
  r"""Monge gap regularizer :cite:`uscidda:23`.

  For a cost function :math:`c` and empirical reference measure
  :math:`\hat{\rho}_n=\frac{1}{n}\sum_{i=1}^n \delta_{x_i}`, the
  (entropic) Monge gap of a map function
  :math:`T:\mathbb{R}^d\rightarrow\mathbb{R}^d` is defined as:

  .. math::
    \mathcal{M}^c_{\hat{\rho}_n, \varepsilon} (T)
    = \frac{1}{n} \sum_{i=1}^n c(x_i, T(x_i)) -
    W_{c, \varepsilon}(\hat{\rho}_n, T \sharp \hat{\rho}_n)

  See :cite:`uscidda:23` Eq. (8). This function is a thin wrapper that calls
  :func:`~ott.neural.methods.monge_gap.monge_gap_from_samples`.

  Args:
    map_fn: Callable corresponding to map :math:`T` in definition above. The
      callable should be vectorized (e.g. using :func:`~jax.vmap`), i.e,
      able to process a *batch* of vectors of size `d`, namely
      ``map_fn`` applied to an array returns an array of the same shape.
    reference_points: Array of `[n,d]` points, :math:`\hat\rho_n`.
    cost_fn: An object of class :class:`~ott.geometry.costs.CostFn`.
    epsilon: Regularization parameter. See
      :class:`~ott.geometry.pointcloud.PointCloud`
    relative_epsilon: when `False`, the parameter ``epsilon`` specifies the
      value of the entropic regularization parameter. When `True`, ``epsilon``
      refers to a fraction of the
      :attr:`~ott.geometry.pointcloud.PointCloud.mean_cost_matrix`, which is
      computed adaptively using ``source`` and ``target`` points.
    scale_cost: option to rescale the cost matrix. Implemented scalings are
      'median', 'mean' and 'max_cost'. Alternatively, a float factor can be
      given to rescale the cost such that ``cost_matrix /= scale_cost``.
    return_output: boolean to also return the
      :class:`~ott.solvers.linear.sinkhorn.SinkhornOutput`.
    kwargs: holds the kwargs to instantiate the or
      :class:`~ott.solvers.linear.sinkhorn.Sinkhorn` solver to
      compute the regularized OT cost.

  Returns:
    The Monge gap value and optionally the
    :class:`~ott.solvers.linear.sinkhorn.SinkhornOutput`
  """
  target = map_fn(reference_points)
  return monge_gap_from_samples(
      source=reference_points,
      target=target,
      cost_fn=cost_fn,
      epsilon=epsilon,
      relative_epsilon=relative_epsilon,
      scale_cost=scale_cost,
      return_output=return_output,
      **kwargs
  )


def monge_gap_from_samples(
    source: jnp.ndarray,
    target: jnp.ndarray,
    cost_fn: Optional[costs.CostFn] = None,
    epsilon: Optional[float] = None,
    relative_epsilon: Optional[bool] = None,
    scale_cost: Union[float, Literal["mean", "max_cost", "median"]] = 1.0,
    return_output: bool = False,
    **kwargs: Any
) -> Union[float, Tuple[float, sinkhorn.SinkhornOutput]]:
  r"""Monge gap, instantiated in terms of samples before / after applying map.

  .. math::
    \frac{1}{n} \sum_{i=1}^n c(x_i, y_i)) -
    W_{c, \varepsilon}(\frac{1}{n}\sum_i \delta_{x_i},
    \frac{1}{n}\sum_i \delta_{y_i})

  where :math:`W_{c, \varepsilon}` is an entropy-regularized optimal transport
  cost, the :attr:`~ott.solvers.linear.sinkhorn.SinkhornOutput.ent_reg_cost`.

  Args:
    source: samples from first measure, array of shape ``[n, d]``.
    target: samples from second measure, array of shape ``[n, d]``.
    cost_fn: a cost function between two points in dimension :math:`d`.
      If :obj:`None`, :class:`~ott.geometry.costs.SqEuclidean` is used.
    epsilon: Regularization parameter. See
      :class:`~ott.geometry.pointcloud.PointCloud`
    relative_epsilon: when `False`, the parameter ``epsilon`` specifies the
      value of the entropic regularization parameter. When `True`, ``epsilon``
      refers to a fraction of the
      :attr:`~ott.geometry.pointcloud.PointCloud.mean_cost_matrix`, which is
      computed adaptively using ``source`` and ``target`` points.
    scale_cost: option to rescale the cost matrix. Implemented scalings are
      'median', 'mean' and 'max_cost'. Alternatively, a float factor can be
      given to rescale the cost such that ``cost_matrix /= scale_cost``.
    return_output: boolean to also return the
      :class:`~ott.solvers.linear.sinkhorn.SinkhornOutput`.
    kwargs: holds the kwargs to instantiate the or
      :class:`~ott.solvers.linear.sinkhorn.Sinkhorn` solver to
      compute the regularized OT cost.

  Returns:
    The Monge gap value and optionally the
    :class:`~ott.solvers.linear.sinkhorn.SinkhornOutput`
  """
  cost_fn = costs.SqEuclidean() if cost_fn is None else cost_fn
  geom = pointcloud.PointCloud(
      x=source,
      y=target,
      cost_fn=cost_fn,
      epsilon=epsilon,
      relative_epsilon=relative_epsilon,
      scale_cost=scale_cost,
  )
  gt_displacement_cost = jnp.mean(jax.vmap(cost_fn)(source, target))
  out = linear.solve(geom=geom, **kwargs)
  loss = gt_displacement_cost - out.ent_reg_cost
  return (loss, out) if return_output else loss


class MongeGapEstimator:
  r"""Mapping estimator between probability measures.

  It estimates a map :math:`T` by minimizing the loss:

  .. math::
    \text{min}_{\theta}\; \Delta(T_\theta \sharp \mu, \theta)
    + \lambda R(T_\theta \sharp \rho, \rho)

  where :math:`\Delta` is a fitting loss and :math:`R` is a regularizer.
  :math:`\Delta` allows to fit the marginal constraint, i.e. transport
  :math:`\mu` to  :math:`\nu` via :math:`T`, while :math:`R`
  is a regularizer imposing an inductive bias on the learned map. The
  regularizer in this case is a function used to compute a metric between two
  sets of points.

  For instance, :math:`\Delta` can be the
  :func:`~ott.tools.sinkhorn_divergence.sinkhorn_divergence`
  and :math:`R` the :func:`~ott.neural.methods.monge_gap.monge_gap_from_samples`
  :cite:`uscidda:23` for a given cost function :math:`c`.
  In that case, it estimates a :math:`c`-OT map, i.e. a map :math:`T`
  optimal for the Monge problem induced by :math:`c`.

  Args:
    dim_data: input dimensionality of data required for network init.
    model: network architecture for map :math:`T`.
    optimizer: optimizer function for map :math:`T`.
    fitting_loss: function that outputs a fitting loss :math:`\Delta` between
      two families of points, as well as any log object.
    regularizer: function that outputs a score from two families of points,
      here assumed to be of the same size, as well as any log object.
    regularizer_strength: strength of the :attr:`regularizer`.
    num_train_iters: number of total training iterations.
    logging: option to return logs.
    valid_freq: frequency with training and validation are logged.
    rng: random key used for seeding for network initializations.
  """

  def __init__(
      self,
      dim_data: int,
      model: potentials.BasePotential,
      optimizer: Optional[optax.OptState] = None,
      fitting_loss: Optional[Callable[[jnp.ndarray, jnp.ndarray],
                                      Tuple[float, Optional[Any]]]] = None,
      regularizer: Optional[Callable[[jnp.ndarray, jnp.ndarray],
                                     Tuple[float, Optional[Any]]]] = None,
      regularizer_strength: Union[float, Sequence[float]] = 1.0,
      num_train_iters: int = 10_000,
      logging: bool = False,
      valid_freq: int = 500,
      rng: Optional[jax.Array] = None,
  ):
    self._fitting_loss = fitting_loss
    self._regularizer = regularizer
    # Can use either a fixed strength, or generalize to a schedule.
    self.regularizer_strength = jnp.repeat(
        jnp.atleast_2d(regularizer_strength),
        num_train_iters,
        total_repeat_length=num_train_iters,
        axis=0
    ).ravel()
    self.num_train_iters = num_train_iters
    self.logging = logging
    self.valid_freq = valid_freq
    self.rng = utils.default_prng_key(rng)

    # set default optimizer
    if optimizer is None:
      optimizer = optax.adam(learning_rate=0.001)

    # setup training
    self.setup(dim_data, model, optimizer)

  def setup(
      self,
      dim_data: int,
      neural_net: potentials.BasePotential,
      optimizer: optax.OptState,
  ):
    """Setup all components required to train the network."""
    # neural network
    self.state_neural_net = neural_net.create_train_state(
        self.rng, optimizer, dim_data
    )

    # step function
    self.step_fn = self._get_step_fn()

  @property
  def regularizer(self) -> Callable[[jnp.ndarray, jnp.ndarray], float]:
    """Regularizer added to the fitting loss.

    Can be, e.g. the :func:`~ott.neural.methods.monge_gap.monge_gap_from_samples`.
    If no regularizer is passed for solver instantiation,
    or regularization weight :attr:`regularizer_strength` is 0,
    return 0 by default along with an empty set of log values.
    """  # noqa: E501
    if self._regularizer is not None:
      return self._regularizer
    return lambda *_, **__: (0.0, None)

  @property
  def fitting_loss(self) -> Callable[[jnp.ndarray, jnp.ndarray], float]:
    """Fitting loss to fit the marginal constraint.

    Can be, e.g. :func:`~ott.tools.sinkhorn_divergence.sinkhorn_divergence`.
    If no fitting_loss is passed for solver instantiation, return 0 by default,
    and no log values.
    """
    if self._fitting_loss is not None:
      return self._fitting_loss
    return lambda *_, **__: (0.0, None)

  @staticmethod
  def _generate_batch(
      loader_source: Iterator[jnp.ndarray],
      loader_target: Iterator[jnp.ndarray],
  ) -> Dict[str, jnp.ndarray]:
    """Generate batches a batch of samples.

    ``loader_source`` and ``loader_target`` can be training or
    validation dataloaders.
    """
    return {
        "source": next(loader_source),
        "target": next(loader_target),
    }

  def train_map_estimator(
      self,
      trainloader_source: Iterator[jnp.ndarray],
      trainloader_target: Iterator[jnp.ndarray],
      validloader_source: Iterator[jnp.ndarray],
      validloader_target: Iterator[jnp.ndarray],
  ) -> Tuple[train_state.TrainState, Dict[str, Any]]:
    """Training loop."""
    # define logs
    logs = collections.defaultdict(lambda: collections.defaultdict(list))

    # try to display training progress with tqdm
    try:
      from tqdm import trange
      tbar = trange(self.num_train_iters, leave=True)
    except ImportError:
      tbar = range(self.num_train_iters)

    for step in tbar:
      #  update step
      is_logging_step = (
          self.logging and ((step % self.valid_freq == 0) or
                            (step == self.num_train_iters - 1))
      )
      train_batch = self._generate_batch(
          loader_source=trainloader_source,
          loader_target=trainloader_target,
      )
      valid_batch = (
          None if not is_logging_step else self._generate_batch(
              loader_source=validloader_source,
              loader_target=validloader_target,
          )
      )
      self.state_neural_net, current_logs = self.step_fn(
          self.state_neural_net, train_batch, valid_batch, is_logging_step, step
      )

      # store and print metrics if logging step
      if is_logging_step:
        for log_key in current_logs:
          for metric_key in current_logs[log_key]:
            logs[log_key][metric_key].append(current_logs[log_key][metric_key])

        # update the tqdm bar if tqdm is available
        if not isinstance(tbar, range):
          reg_msg = (
              "NA" if current_logs["eval"]["regularizer"] == 0.0 else
              f"{current_logs['eval']['regularizer']:.4f}"
          )
          postfix_str = (
              f"fitting_loss: {current_logs['eval']['fitting_loss']:.4f}, "
              f"regularizer: {reg_msg} ,"
              f"total: {current_logs['eval']['total_loss']:.4f}"
          )
          tbar.set_postfix_str(postfix_str)

    return self.state_neural_net, logs

  def _get_step_fn(self) -> Callable:
    """Create a one step training and evaluation function."""

    def loss_fn(
        params: frozen_dict.FrozenDict, apply_fn: Callable,
        batch: Dict[str, jnp.ndarray], step: int
    ) -> Tuple[float, Dict[str, float]]:
      """Loss function."""
      # map samples with the fitted map
      mapped_samples = apply_fn({"params": params}, batch["source"])

      # compute the loss
      val_fitting_loss, log_fitting_loss = self.fitting_loss(
          mapped_samples, batch["target"]
      )
      val_regularizer, log_regularizer = self.regularizer(
          batch["source"], mapped_samples
      )
      val_tot_loss = (
          val_fitting_loss + self.regularizer_strength[step] * val_regularizer
      )

      # store training logs
      loss_logs = {
          "total_loss": val_tot_loss,
          "fitting_loss": val_fitting_loss,
          "regularizer": val_regularizer,
          "log_regularizer": log_regularizer,
          "log_fitting": log_fitting_loss,
      }

      return val_tot_loss, loss_logs

    @functools.partial(jax.jit, static_argnums=3)
    def step_fn(
        state_neural_net: train_state.TrainState,
        train_batch: Dict[str, jnp.ndarray],
        valid_batch: Optional[Dict[str, jnp.ndarray]] = None,
        is_logging_step: bool = False,
        step: int = 0
    ) -> Tuple[train_state.TrainState, Dict[str, float]]:
      """One step function."""
      # compute loss and gradients
      grad_fn = jax.value_and_grad(loss_fn, argnums=0, has_aux=True)
      (_, current_train_logs), grads = grad_fn(
          state_neural_net.params, state_neural_net.apply_fn, train_batch, step
      )

      # logging step
      current_logs = {"train": current_train_logs, "eval": {}}
      if is_logging_step:
        _, current_eval_logs = loss_fn(
            params=state_neural_net.params,
            apply_fn=state_neural_net.apply_fn,
            batch=valid_batch,
            step=step
        )
        current_logs["eval"] = current_eval_logs

      # update state
      return state_neural_net.apply_gradients(grads=grads), current_logs

    return step_fn



# Copyright OTT-JAX
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#   http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from typing import Any, Optional

import jax
import jax.numpy as jnp

__all__ = ["Epsilon", "DEFAULT_SCALE"]

#: Scaling applied to statistic (mean/std) of cost to compute default epsilon.
DEFAULT_SCALE = 0.05


@jax.tree_util.register_pytree_node_class
class Epsilon:
  """Scheduler class for the regularization parameter epsilon.

  An epsilon scheduler outputs a regularization strength, to be used by in a
  Sinkhorn-type algorithm, at any iteration count. That value is either the
  final, targeted regularization, or one that is larger, obtained by
  geometric decay of an initial value that is larger than the intended target.
  Concretely, the value returned by such a scheduler will consider first
  the max between ``target`` and ``init * target * decay ** iteration``.
  If the ``scale_epsilon`` parameter is provided, that value is used to
  multiply the max computed previously by ``scale_epsilon``.

  Args:
    target: the epsilon regularizer that is targeted. If :obj:`None`,
      use :obj:`DEFAULT_SCALE`, currently set at :math:`0.05`.
    scale_epsilon: if passed, used to multiply the regularizer, to rescale it.
      If :obj:`None`, use :math:`1`.
    init: initial value when using epsilon scheduling, understood as multiple
      of target value. if passed, ``int * decay ** iteration`` will be used
      to rescale target.
    decay: geometric decay factor, :math:`<1`.
  """

  def __init__(
      self,
      target: Optional[float] = None,
      scale_epsilon: Optional[float] = None,
      init: float = 1.0,
      decay: float = 1.0
  ):
    self._target_init = target
    self._scale_epsilon = scale_epsilon
    self._init = init
    self._decay = decay

  @property
  def target(self) -> float:
    """Return the final regularizer value of scheduler."""
    target = DEFAULT_SCALE if self._target_init is None else self._target_init
    scale = 1.0 if self._scale_epsilon is None else self._scale_epsilon
    return scale * target

  def at(self, iteration: Optional[int] = 1) -> float:
    """Return (intermediate) regularizer value at a given iteration."""
    if iteration is None:
      return self.target
    # check the decay is smaller than 1.0.
    decay = jnp.minimum(self._decay, 1.0)
    # the multiple is either 1.0 or a larger init value that is decayed.
    multiple = jnp.maximum(self._init * (decay ** iteration), 1.0)
    return multiple * self.target

  def done(self, eps: float) -> bool:
    """Return whether the scheduler is done at a given value."""
    return eps == self.target

  def done_at(self, iteration: Optional[int]) -> bool:
    """Return whether the scheduler is done at a given iteration."""
    return self.done(self.at(iteration))

  def set(self, **kwargs: Any) -> "Epsilon":
    """Return a copy of self, with potential overwrites."""
    kwargs = {
        "target": self._target_init,
        "scale_epsilon": self._scale_epsilon,
        "init": self._init,
        "decay": self._decay,
        **kwargs
    }
    return Epsilon(**kwargs)

  def tree_flatten(self):  # noqa: D102
    return (
        self._target_init, self._scale_epsilon, self._init, self._decay
    ), None

  @classmethod
  def tree_unflatten(cls, aux_data, children):  # noqa: D102
    del aux_data
    return cls(*children)
